/*
 *  PCM Interface - main file
 *  Copyright (c) 1998 by Jaroslav Kysela <perex@suse.cz>
 *
 *
 *   This library is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Library General Public License as
 *   published by the Free Software Foundation; either version 2 of
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU Library General Public License for more details.
 *
 *   You should have received a copy of the GNU Library General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <sys/pathmgr.h>
#include <libgen.h>
#include <sys/utsname.h>
#include "pcm_local.h"
#include <devctl.h>

__SRCVERSION( "$URL: http://svn.ott.qnx.com/product/branches/6.5.0/SP1/lib/asound/pcm/pcm.c $ $Rev: 646317 $" );

#define SND_FILE_PCM_PATH				"/dev/snd/"
#define SND_FILE_PCM_PLAYBACK			"/dev/snd/pcmC%iD%ip"
#define SND_FILE_PCM_CAPTURE			"/dev/snd/pcmC%iD%ic"
#define SND_FILE_PCM_PREFERRED_PLAYBACK	"/dev/snd/pcmPreferredp"
#define SND_FILE_PCM_PREFERRED_CAPTURE	"/dev/snd/pcmPreferredc"
#define SND_FILE_CFG_PREFERENCES		"/etc/config/trap/audio/%s-preferences"


int
snd_pcm_find (unsigned int format, int *number, int *cards, int *devices, int mode)
{
	char   *filefmt;
	int     index;
	int     max_cards, card;
	snd_ctl_t *handle;
	int     max_devs, dev;
	char    filename[32];
	int     fd;
	snd_ctl_hw_info_t hw_info;
	snd_pcm_channel_info_t info;

	for (index = 0; index < *number; index++)
	{
		cards[index] = -1;
		devices[index] = -1;
	}

	if (mode & SND_PCM_OPEN_CAPTURE && mode & SND_PCM_OPEN_PLAYBACK)
		return -EINVAL;
	switch (mode)
	{
	case SND_PCM_CHANNEL_PLAYBACK:
		filefmt = SND_FILE_PCM_PLAYBACK;
		break;
	case SND_PCM_CHANNEL_CAPTURE:
		filefmt = SND_FILE_PCM_CAPTURE;
		break;
	default:
		return -EINVAL;
	}


	index = 0;
	max_cards = snd_cards ();
	for (card = 0; card < max_cards; card++)
	{
		if (snd_ctl_open (&handle, card) < 0)
			continue;

		if (snd_ctl_hw_info (handle, &hw_info) < 0)
			continue;

		max_devs = hw_info.pcmdevs;
		for (dev = 0; dev < max_devs; dev++)
		{
			memset (&info, 0, sizeof (info));

			if (snd_ctl_driver_version (handle) >= 3)
			{
				if (snd_ctl_pcm_channel_info (handle, dev, mode, 0, &info) < 0)
					continue;
			}
			else
			{
				sprintf (filename, filefmt, card, dev);
				if ((fd = open (filename, O_RDWR | O_NONBLOCK | O_CLOEXEC)) == -1)
				{
					close (fd);
					continue;
				}
				if (ioctl (fd, SND_PCM_IOCTL_CHANNEL_INFO, &info) < 0)
				{
					close (fd);
					continue;
				}
				close (fd);
			}

			if ((info.formats & format))
			{
				cards[index] = card;
				devices[index] = dev;
				index++;
			}
			if (index == *number)
				break;
		}

		snd_ctl_close (handle);
	}

	return (*number = index);
}


static int
snd_pcm_open_internal (snd_pcm_t ** pcm, char *name, int *rcard, int *rdevice, int mode)
{
	int     fd;
	int     ver;
	snd_pcm_info_t pinfo;

	if ((*pcm) == NULL)
	{
		if (((*pcm) = (snd_pcm_t *) calloc (1, sizeof (snd_pcm_t))) == NULL)
		{
			return -errno;
		}
		(*pcm)->fd[SND_PCM_CHANNEL_PLAYBACK] = -1;
		(*pcm)->fd[SND_PCM_CHANNEL_CAPTURE] = -1;
	}

	if ((fd = open (name, O_RDWR | O_CLOEXEC |
				((mode & SND_PCM_OPEN_NONBLOCK) ? O_NONBLOCK : 0))) != -1)
	{
		if (ioctl (fd, SND_PCM_IOCTL_PVERSION, &ver) < 0)
		{
			close (fd);
			free (*pcm);
			return -errno;
		}
		if (SND_PROTOCOL_INCOMPATIBLE (ver, SND_PCM_VERSION))
		{
			close (fd);
			free (*pcm);
			return -SND_ERROR_INCOMPATIBLE_VERSION;
		}
		if (ioctl (fd, SND_PCM_IOCTL_INFO, &pinfo) < 0)
		{
			close (fd);
			free (*pcm);
			return -errno;
		}
		(*pcm)->card = pinfo.card;
		(*pcm)->device = pinfo.device;

		if (mode & SND_PCM_OPEN_PLAYBACK)
			(*pcm)->fd[SND_PCM_CHANNEL_PLAYBACK] = fd;
		else
			(*pcm)->fd[SND_PCM_CHANNEL_CAPTURE] = fd;
		(*pcm)->mode = mode;
		(*pcm)->ver = ver;
	}
	else
	{
		free (*pcm);
		return -ENOENT;
	}

	if (rcard != NULL)
		*rcard = (*pcm)->card;
	if (rdevice != NULL)
		*rdevice = (*pcm)->device;
	return 0;
}

int
snd_pcm_open (snd_pcm_t ** handle, int card, int device, int mode)
{
	char    name[32];
	int     rtn;

	*handle = NULL;
	if ((mode & (SND_PCM_OPEN_PLAYBACK | SND_PCM_OPEN_CAPTURE)) == 0)
		return -EINVAL;
	if (mode & SND_PCM_OPEN_PLAYBACK)
	{
		sprintf (name, SND_FILE_PCM_PLAYBACK, card, device);
		if ((rtn = snd_pcm_open_internal (handle, name, NULL, NULL,
					(mode & ~SND_PCM_OPEN_CAPTURE))) != 0)
			return rtn;
	}
	if (mode & SND_PCM_OPEN_CAPTURE)
	{
		sprintf (name, SND_FILE_PCM_CAPTURE, card, device);
		if ((rtn = snd_pcm_open_internal (handle, name, NULL, NULL,
					(mode & ~SND_PCM_OPEN_PLAYBACK))) != 0)
			return rtn;
	}
	return 0;
}

int
snd_pcm_open_name (snd_pcm_t ** handle, char *name, int mode)
{
	char    path[_POSIX_PATH_MAX];

	*handle = NULL;
	if ((mode & (SND_PCM_OPEN_PLAYBACK | SND_PCM_OPEN_CAPTURE)) == 0)
		return -EINVAL;
	if (mode & SND_PCM_OPEN_CAPTURE && mode & SND_PCM_OPEN_PLAYBACK)
		return -EINVAL;
	if (name[0] != '/')
	{
		strcpy (path, SND_FILE_PCM_PATH);
		strcat (path, name);
		name = path;
	}
	return (snd_pcm_open_internal (handle, name, NULL, NULL, mode));
}

int
snd_pcm_open_preferred (snd_pcm_t ** handle, int *rcard, int *rdevice, int mode)
{
	char   *name;

	*handle = NULL;
	if (mode & SND_PCM_OPEN_CAPTURE && mode & SND_PCM_OPEN_PLAYBACK)
		return -EINVAL;
	else if (mode & SND_PCM_OPEN_PLAYBACK)
		name = SND_FILE_PCM_PREFERRED_PLAYBACK;
	else if (mode & SND_PCM_OPEN_CAPTURE)
		name = SND_FILE_PCM_PREFERRED_CAPTURE;
	else
		return -EINVAL;
	return (snd_pcm_open_internal (handle, name, rcard, rdevice, mode));
}

int
snd_pcm_close (snd_pcm_t * pcm)
{
	int     res = EOK;
	int     channel;

	if (!pcm)
		return -EINVAL;
	for (channel = 0; channel < 2; ++channel)
	{
		snd_pcm_munmap (pcm, channel);
		snd_pcm_plugin_clear (pcm, channel);
		if (pcm->fd[channel] >= 0)
			if (close (pcm->fd[channel]))
				res = -errno;
	}
	free (pcm);
	pcm = NULL;
	return res;
}

int
snd_pcm_file_descriptor (snd_pcm_t * pcm, int channel)
{
	if (!pcm)
		return -EINVAL;
	if (channel < 0 || channel > 1)
		return -EINVAL;
	return pcm->fd[channel];
}

int
snd_pcm_nonblock_mode (snd_pcm_t * pcm, int nonblock)
{
	long    flags;
	int     fd, channel;

	if (!pcm)
		return -EINVAL;
	for (channel = 0; channel < 2; ++channel)
	{
		fd = pcm->fd[channel];
		if (fd < 0)
			continue;
		if ((flags = fcntl (fd, F_GETFL)) < 0)
			return -errno;
		if (nonblock)
			flags |= O_NONBLOCK;
		else
			flags &= ~O_NONBLOCK;
		if (fcntl (fd, F_SETFL, flags) < 0)
			return -errno;
		if (nonblock)
			pcm->mode |= SND_PCM_OPEN_NONBLOCK;
		else
			pcm->mode &= ~SND_PCM_OPEN_NONBLOCK;
	}
	return 0;
}

int
snd_pcm_info (snd_pcm_t * pcm, snd_pcm_info_t * info)
{
	int     fd, channel;

	if (!pcm || !info)
		return -EINVAL;
	for (channel = 0; channel < 2; ++channel)
	{
		fd = pcm->fd[channel];
		if (fd >= 0)
			break;
	}
	if (ioctl (fd, SND_PCM_IOCTL_INFO, info) < 0)
		return -errno;
	return 0;
}

int
snd_pcm_channel_info (snd_pcm_t * pcm, snd_pcm_channel_info_t * info)
{
	int     fd;

	if (!pcm || !info)
		return -EINVAL;
	if (info->channel < 0 || info->channel > 1)
		return -EINVAL;
	fd = pcm->fd[info->channel];
	if (fd < 0)
		return -EINVAL;
	if (ioctl (fd, SND_PCM_IOCTL_CHANNEL_INFO, info) < 0)
		return -errno;
	return 0;
}

int
snd_pcm_channel_params (snd_pcm_t * pcm, snd_pcm_channel_params_t * params)
{
	int     err;
	int     fd;

	if (!pcm || !params)
		return -EINVAL;
	if (params->channel < 0 || params->channel > 1)
		return -EINVAL;
	fd = pcm->fd[params->channel];
	if (fd < 0)
		return -EINVAL;
	if (ioctl (fd, SND_PCM_IOCTL_CHANNEL_PARAMS, params) < 0)
		return -errno;
	pcm->setup_is_valid[params->channel] = 0;
	memset (&pcm->setup[params->channel], 0, sizeof (snd_pcm_channel_setup_t));
	pcm->setup[params->channel].channel = params->channel;
	if ((err = snd_pcm_channel_setup (pcm, &pcm->setup[params->channel])) < 0)
		return err;
	pcm->setup_is_valid[params->channel] = 1;
	return 0;
}

int
snd_pcm_channel_setup (snd_pcm_t * pcm, snd_pcm_channel_setup_t * setup)
{
	int     fd;
	iov_t   iov[3];
	int     rtn;
	snd_mixer_eid_t *eid_ptr;
	snd_mixer_gid_t *gid_ptr;

	if (!pcm || !setup)
		return -EINVAL;
	if (setup->channel < 0 || setup->channel > 1)
		return -EINVAL;
	fd = pcm->fd[setup->channel];
	if (fd < 0)
		return -EINVAL;
	if (pcm->setup_is_valid[setup->channel] == 0)
	{
		SETIOV (&iov[0], &pcm->setup[setup->channel], sizeof (snd_pcm_channel_setup_t));
		SETIOV (&iov[1], &pcm->setup_eid[setup->channel], sizeof (snd_mixer_eid_t));
		SETIOV (&iov[2], &pcm->setup_gid[setup->channel], sizeof (snd_mixer_gid_t));
		if ((rtn = devctlv (fd, SND_PCM_IOCTL_CHANNEL_SETUP, 3, 3, iov, iov, NULL)) != EOK)
			return -rtn;
		pcm->setup_is_valid[setup->channel] = 1;
	}
	eid_ptr = setup->mixer_eid;
	gid_ptr = setup->mixer_gid;
	memcpy (setup, &pcm->setup[setup->channel], sizeof (*setup));
	if ((setup->mixer_eid = eid_ptr))
		memcpy (setup->mixer_eid, &pcm->setup_eid[setup->channel], sizeof (snd_mixer_eid_t));
	if ((setup->mixer_gid = gid_ptr))
		memcpy (setup->mixer_gid, &pcm->setup_gid[setup->channel], sizeof (snd_mixer_gid_t));
	return 0;
}

int
snd_pcm_channel_status (snd_pcm_t * pcm, snd_pcm_channel_status_t * status)
{
	int     fd;

	if (!pcm || !status)
		return -EINVAL;
	fd = pcm->fd[status->channel];
	if (fd < 0)
		return -EINVAL;
	if (ioctl (fd, SND_PCM_IOCTL_CHANNEL_STATUS, status) < 0)
		return -errno;
	status->subbuffered = 0;
	return 0;
}

int
snd_pcm_playback_prepare (snd_pcm_t * pcm)
{
	if (!pcm)
		return -EINVAL;
	if (pcm->fd[SND_PCM_CHANNEL_PLAYBACK] < 0)
		return -EINVAL;
	if (ioctl (pcm->fd[SND_PCM_CHANNEL_PLAYBACK], SND_PCM_IOCTL_CHANNEL_PREPARE) < 0)
		return -errno;
	return 0;
}

int
snd_pcm_capture_prepare (snd_pcm_t * pcm)
{
	if (!pcm)
		return -EINVAL;
	if (pcm->fd[SND_PCM_CHANNEL_CAPTURE] < 0)
		return -EINVAL;
	if (ioctl (pcm->fd[SND_PCM_CHANNEL_CAPTURE], SND_PCM_IOCTL_CHANNEL_PREPARE) < 0)
		return -errno;
	return 0;
}

int
snd_pcm_channel_prepare (snd_pcm_t * pcm, int channel)
{
	switch (channel)
	{
	case SND_PCM_CHANNEL_PLAYBACK:
		return snd_pcm_playback_prepare (pcm);
	case SND_PCM_CHANNEL_CAPTURE:
		return snd_pcm_capture_prepare (pcm);
	default:
		return -EIO;
	}
}

int
snd_pcm_playback_go (snd_pcm_t * pcm)
{
	if (!pcm)
		return -EINVAL;
	if (pcm->fd[SND_PCM_CHANNEL_PLAYBACK] < 0)
		return -EINVAL;
	if (ioctl (pcm->fd[SND_PCM_CHANNEL_PLAYBACK], SND_PCM_IOCTL_CHANNEL_GO) < 0)
		return -errno;
	return 0;
}

int
snd_pcm_capture_go (snd_pcm_t * pcm)
{
	if (!pcm)
		return -EINVAL;
	if (pcm->fd[SND_PCM_CHANNEL_CAPTURE] < 0)
		return -EINVAL;
	if (ioctl (pcm->fd[SND_PCM_CHANNEL_CAPTURE], SND_PCM_IOCTL_CHANNEL_GO) < 0)
		return -errno;
	return 0;
}

int
snd_pcm_channel_go (snd_pcm_t * pcm, int channel)
{
	switch (channel)
	{
	case SND_PCM_CHANNEL_PLAYBACK:
		return snd_pcm_playback_go (pcm);
	case SND_PCM_CHANNEL_CAPTURE:
		return snd_pcm_capture_go (pcm);
	default:
		return -EIO;
	}
}

int
snd_pcm_playback_pause (snd_pcm_t * pcm)
{
	int     on = 1;

	if (!pcm)
		return -EINVAL;
	if (pcm->fd[SND_PCM_CHANNEL_PLAYBACK] < 0)
		return -EINVAL;
	if (ioctl (pcm->fd[SND_PCM_CHANNEL_PLAYBACK], SND_PCM_IOCTL_CHANNEL_PAUSE, &on) < 0)
		return -errno;
	return 0;
}

int
snd_pcm_capture_pause (snd_pcm_t * pcm)
{
	int     on = 1;

	if (!pcm)
		return -EINVAL;
	if (pcm->fd[SND_PCM_CHANNEL_CAPTURE] < 0)
		return -EINVAL;
	if (ioctl (pcm->fd[SND_PCM_CHANNEL_CAPTURE], SND_PCM_IOCTL_CHANNEL_PAUSE, &on) < 0)
		return -errno;
	return 0;
}

int
snd_pcm_channel_pause (snd_pcm_t * pcm, int channel)
{
	switch (channel)
	{
	case SND_PCM_CHANNEL_PLAYBACK:
		return snd_pcm_playback_pause (pcm);
	case SND_PCM_CHANNEL_CAPTURE:
		return snd_pcm_capture_pause (pcm);
	default:
		return -EIO;
	}
}

int
snd_pcm_playback_resume (snd_pcm_t * pcm)
{
	int     off = 0;

	if (!pcm)
		return -EINVAL;
	if (pcm->fd[SND_PCM_CHANNEL_PLAYBACK] < 0)
		return -EINVAL;
	if (ioctl (pcm->fd[SND_PCM_CHANNEL_PLAYBACK], SND_PCM_IOCTL_CHANNEL_PAUSE, &off) < 0)
		return -errno;
	return 0;
}

int
snd_pcm_capture_resume (snd_pcm_t * pcm)
{
	int     off = 0;

	if (!pcm)
		return -EINVAL;
	if (pcm->fd[SND_PCM_CHANNEL_CAPTURE] < 0)
		return -EINVAL;
	if (ioctl (pcm->fd[SND_PCM_CHANNEL_CAPTURE], SND_PCM_IOCTL_CHANNEL_PAUSE, &off) < 0)
		return -errno;
	return 0;
}

int
snd_pcm_channel_resume (snd_pcm_t * pcm, int channel)
{
	switch (channel)
	{
	case SND_PCM_CHANNEL_PLAYBACK:
		return snd_pcm_playback_resume (pcm);
	case SND_PCM_CHANNEL_CAPTURE:
		return snd_pcm_capture_resume (pcm);
	default:
		return -EIO;
	}
}

int
snd_pcm_playback_drain (snd_pcm_t * pcm)
{
	if (!pcm)
		return -EINVAL;
	if (pcm->fd[SND_PCM_CHANNEL_PLAYBACK] < 0)
		return -EINVAL;
	if (ioctl (pcm->fd[SND_PCM_CHANNEL_PLAYBACK], SND_PCM_IOCTL_CHANNEL_DRAIN) < 0)
		return -errno;
	return 0;
}

int
snd_pcm_playback_flush (snd_pcm_t * pcm)
{
	if (!pcm)
		return -EINVAL;
	if (pcm->fd[SND_PCM_CHANNEL_PLAYBACK] < 0)
		return -EINVAL;
	if (ioctl (pcm->fd[SND_PCM_CHANNEL_PLAYBACK], SND_PCM_IOCTL_CHANNEL_FLUSH) < 0)
		return -errno;
	return 0;
}

int
snd_pcm_capture_flush (snd_pcm_t * pcm)
{
	if (!pcm)
		return -EINVAL;
	if (pcm->fd[SND_PCM_CHANNEL_CAPTURE] < 0)
		return -EINVAL;
	if (ioctl (pcm->fd[SND_PCM_CHANNEL_CAPTURE], SND_PCM_IOCTL_CHANNEL_FLUSH) < 0)
		return -errno;
	return 0;
}

int
snd_pcm_channel_flush (snd_pcm_t * pcm, int channel)
{
	switch (channel)
	{
	case SND_PCM_CHANNEL_PLAYBACK:
		return snd_pcm_playback_flush (pcm);
	case SND_PCM_CHANNEL_CAPTURE:
		return snd_pcm_capture_flush (pcm);
	default:
		return -EIO;
	}
}

ssize_t
snd_pcm_transfer_size (snd_pcm_t * pcm, int channel)
{
	if (!pcm || channel < 0 || channel > 1)
		return -EINVAL;
	if (!pcm->setup_is_valid[channel])
		return -EBADFD;
	if (pcm->setup[channel].mode != SND_PCM_MODE_BLOCK)
		return -EBADFD;
	return pcm->setup[channel].buf.block.frag_size;
}

ssize_t
snd_pcm_write (snd_pcm_t * pcm, const void *buffer, size_t size)
{
	ssize_t result;

	if (!pcm || (!buffer && size > 0) || size < 0)
		return -EINVAL;
	if (pcm->fd[SND_PCM_CHANNEL_PLAYBACK] < 0)
		return -EINVAL;
	result = write (pcm->fd[SND_PCM_CHANNEL_PLAYBACK], buffer, size);
	if (result < 0)
		return -errno;
	return result;
}

ssize_t
snd_pcm_read (snd_pcm_t * pcm, void *buffer, size_t size)
{
	ssize_t result;

	if (!pcm || (!buffer && size > 0) || size < 0)
		return -EINVAL;
	if (pcm->fd[SND_PCM_CHANNEL_CAPTURE] < 0)
		return -EINVAL;
	result = read (pcm->fd[SND_PCM_CHANNEL_CAPTURE], buffer, size);
	if (result < 0)
		return -errno;
	return result;
}

int
snd_pcm_mmap (snd_pcm_t * pcm, int channel, snd_pcm_mmap_control_t ** control, void **buffer)
{
	snd_pcm_mmap_info_t info;
	int     fd, err;
	void   *caddr, *daddr;

	if (control)
		*control = NULL;
	if (buffer)
		*buffer = NULL;
	if (!pcm || channel < 0 || channel > 1 || !control || !buffer)
		return -EINVAL;
	fd = pcm->fd[channel];
	if (fd < 0)
		return -EINVAL;
	if (ioctl (fd, SND_PCM_IOCTL_MMAP_INFO, &info) < 0)
		return -errno;

	if ((fd = shm_open (info.dmactl_name, O_RDWR | O_CLOEXEC, 0)) == -1)
		return -errno;
	if ((caddr =
			mmap (NULL, info.ctl_size, PROT_READ | PROT_WRITE | PROT_NOCACHE, MAP_SHARED, fd,
				0)) == MAP_FAILED)
	{
		close (fd);
		return -errno;
	}
	close (fd);
	if ((fd = shm_open (info.dmabuf_name, O_RDWR | O_CLOEXEC, 0)) == -1)
		return -errno;
	if ((daddr =
			mmap (NULL, info.size, PROT_READ | PROT_WRITE | PROT_NOCACHE, MAP_SHARED, fd,
				0)) == MAP_FAILED)
	{
		close (fd);
		err = -errno;
		munmap (caddr, info.ctl_size);
		return err;
	}
	close (fd);
	*control = pcm->mmap_caddr[channel] = caddr;
	*buffer = pcm->mmap_daddr[channel] = daddr;
	pcm->mmap_size[channel] = info.size;
	return 0;
}

int
snd_pcm_munmap (snd_pcm_t * pcm, int channel)
{
	if (!pcm || channel < 0 || channel > 1)
		return -EINVAL;
	if (pcm->mmap_caddr[channel])
	{
		int     frags = pcm->mmap_caddr[channel]->status.frags;

		munmap (pcm->mmap_caddr[channel],
			sizeof (snd_pcm_mmap_control_t) + (sizeof (snd_pcm_mmap_fragment_t) * frags));
		pcm->mmap_caddr[channel] = NULL;
	}
	if (pcm->mmap_daddr[channel])
	{
		munmap (pcm->mmap_daddr[channel], pcm->mmap_size[channel]);
		pcm->mmap_daddr[channel] = NULL;
		pcm->mmap_size[channel] = 0;
	}
	return 0;
}
