/*
 *  PCM Plug-In Interface
 *  Copyright (c) 1999 by Jaroslav Kysela <perex@suse.cz>
 *
 *
 *   This library is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Library General Public License as
 *   published by the Free Software Foundation; either version 2 of
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU Library General Public License for more details.
 *
 *   You should have received a copy of the GNU Library General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <math.h>
#include <sys/ioctl.h>
#include "pcm_local.h"

__SRCVERSION( "$URL: http://svn.ott.qnx.com/product/branches/6.5.0/SP1/lib/asound/pcm/pcm_plugin.c $ $Rev: 229712 $" );

static ssize_t snd_pcm_plugin_write1 (snd_pcm_t * pcm, const void *buffer, size_t count);


unsigned char
snd_pcm_plugin_silence (snd_pcm_format_t * format)
{
	switch (format->format)
	{
	case SND_PCM_SFMT_IMA_ADPCM:						   /* special case */
	case SND_PCM_SFMT_MPEG:
	case SND_PCM_SFMT_GSM:
	case SND_PCM_SFMT_MU_LAW:
	case SND_PCM_SFMT_A_LAW:
		return 0;
	case SND_PCM_SFMT_U8:
	case SND_PCM_SFMT_U16_LE:
	case SND_PCM_SFMT_U16_BE:
	case SND_PCM_SFMT_U24_LE:
	case SND_PCM_SFMT_U24_BE:
	case SND_PCM_SFMT_U32_LE:
	case SND_PCM_SFMT_U32_BE:
		return 0x80;
	case SND_PCM_SFMT_S8:
	case SND_PCM_SFMT_S16_LE:
	case SND_PCM_SFMT_S16_BE:
	case SND_PCM_SFMT_S24_LE:
	case SND_PCM_SFMT_S24_BE:
	case SND_PCM_SFMT_S32_LE:
	case SND_PCM_SFMT_S32_BE:
		return 0;
	case SND_PCM_SFMT_FLOAT:
	case SND_PCM_SFMT_FLOAT64:
	case SND_PCM_SFMT_IEC958_SUBFRAME_LE:
	case SND_PCM_SFMT_IEC958_SUBFRAME_BE:
		return 0;
	}
	return 0;
}

snd_pcm_plugin_t *
snd_pcm_plugin_build (const char *name, int extra)
{
	snd_pcm_plugin_t *plugin;

	if (extra < 0)
		return NULL;
	plugin = (snd_pcm_plugin_t *) calloc (1, sizeof (*plugin) + extra);
	if (plugin == NULL)
		return NULL;
	plugin->name = name ? strdup (name) : NULL;
	return plugin;
}

int
snd_pcm_plugin_free (snd_pcm_plugin_t * plugin)
{
	if (plugin)
	{
		if (plugin->private_free)
			plugin->private_free (plugin, plugin->private_data);
		if (plugin->name)
			free (plugin->name);
		free (plugin);
	}
	return 0;
}

int
snd_pcm_plugin_clear (snd_pcm_t * pcm, int channel)
{
	snd_pcm_plugin_t *plugin, *plugin_next;
	int     idx;

	if (!pcm)
		return -EINVAL;
	plugin = pcm->plugin_first[channel];
	pcm->plugin_first[channel] = NULL;
	pcm->plugin_last[channel] = NULL;
	while (plugin)
	{
		plugin_next = plugin->next;
		snd_pcm_plugin_free (plugin);
		plugin = plugin_next;
	}
	for (idx = 0; idx < 4; idx++)
	{
		if (pcm->plugin_alloc_ptr[idx])
			free (pcm->plugin_alloc_ptr[idx]);
		pcm->plugin_alloc_ptr[idx] = 0;
		pcm->plugin_alloc_size[idx] = 0;
		pcm->plugin_alloc_lock[idx] = 0;
	}
	if (pcm->plugin_alloc_xptr[channel])
		free (pcm->plugin_alloc_xptr[channel]);
	pcm->plugin_alloc_xptr[channel] = NULL;
	pcm->plugin_alloc_xsize[channel] = 0;

	pcm->plugin_partial_block_size[channel] = 0;
	if (pcm->plugin_partial_block_buffer[channel])
		free (pcm->plugin_partial_block_buffer[channel]);
	pcm->plugin_partial_block_buffer[channel] = NULL;
	pcm->plugin_partial_block_len[channel] = 0;
	return 0;
}

int
snd_pcm_plugin_insert (snd_pcm_t * pcm, int channel, snd_pcm_plugin_t * plugin)
{
	if (!pcm || channel < 0 || channel > 1 || !plugin)
		return -EINVAL;
	plugin->next = pcm->plugin_first[channel];
	plugin->prev = NULL;
	if (pcm->plugin_first[channel])
	{
		pcm->plugin_first[channel]->prev = plugin;
		pcm->plugin_first[channel] = plugin;
	}
	else
	{
		pcm->plugin_last[channel] = pcm->plugin_first[channel] = plugin;
	}
	return 0;
}

int
snd_pcm_plugin_append (snd_pcm_t * pcm, int channel, snd_pcm_plugin_t * plugin)
{
	if (!pcm || channel < 0 || channel > 1 || !plugin)
		return -EINVAL;
	plugin->next = NULL;
	plugin->prev = pcm->plugin_last[channel];
	if (pcm->plugin_last[channel])
	{
		pcm->plugin_last[channel]->next = plugin;
		pcm->plugin_last[channel] = plugin;
	}
	else
	{
		pcm->plugin_last[channel] = pcm->plugin_first[channel] = plugin;
	}
	return 0;
}

int
snd_pcm_plugin_remove_to (snd_pcm_t * pcm, int channel, snd_pcm_plugin_t * plugin)
{
	snd_pcm_plugin_t *plugin1, *plugin1_prev;

	if (!pcm || channel < 0 || channel > 1 || !plugin || !plugin->prev)
		return -EINVAL;
	plugin1 = plugin;
	while (plugin1->prev)
		plugin1 = plugin1->prev;
	if (pcm->plugin_first[channel] != plugin1)
		return -EINVAL;
	pcm->plugin_first[channel] = plugin;
	plugin1 = plugin->prev;
	plugin->prev = NULL;
	while (plugin1)
	{
		plugin1_prev = plugin1->prev;
		snd_pcm_plugin_free (plugin1);
		plugin1 = plugin1_prev;
	}
	return 0;
}

snd_pcm_plugin_t *
snd_pcm_plugin_first (snd_pcm_t * pcm, int channel)
{
	if (!pcm || channel < 0 || channel > 1)
		return NULL;
	return pcm->plugin_first[channel];
}

snd_pcm_plugin_t *
snd_pcm_plugin_last (snd_pcm_t * pcm, int channel)
{
	if (!pcm || channel < 0 || channel > 1)
		return NULL;
	return pcm->plugin_last[channel];
}

snd_pcm_plugin_t *
snd_pcm_plugin_find (snd_pcm_t * pcm, int channel, char * plugin_name)
{
	snd_pcm_plugin_t *plugin;

	plugin = pcm->plugin_first[channel];
	while (plugin)
	{
		if (strcmp(plugin_name, plugin->name) == 0)
		{
			return(plugin);
		}
		plugin = plugin->next;
	}
	return NULL;
}

int
snd_pcm_plugin_info (snd_pcm_t * pcm, snd_pcm_channel_info_t * info)
{
	int     err;

	if ((err = snd_pcm_channel_info (pcm, info)) < 0)
		return err;
	info->formats = snd_pcm_plugin_formats (info->formats);
	info->min_rate = 4000;
	info->max_rate = 192000;
	info->rates = SND_PCM_RATE_8000_48000;
	return 0;
}

static int
snd_pcm_plugin_action (snd_pcm_t * pcm, int channel, int action)
{
	snd_pcm_plugin_t *plugin;
	int     err;

	plugin = pcm->plugin_first[channel];
	while (plugin)
	{
		if (plugin->action)
		{
			if ((err = plugin->action (plugin, action)) < 0)
				return err;
		}
		plugin = plugin->next;
	}
	return 0;
}

unsigned int
snd_pcm_plugin_set_disable (snd_pcm_t * pcm, unsigned int mask)
{
	return (pcm->plugin_disable_mask = mask);
}

int
snd_pcm_plugin_params (snd_pcm_t * pcm, snd_pcm_channel_params_t * params)
{
	int     sample_size;
	snd_pcm_channel_params_t hwparams;
	snd_pcm_plugin_t *plugin;
	int     err;

	if (!pcm || !params || params->channel < 0 || params->channel > 1 ||
		pcm->fd[params->channel] < 0)
		return -EINVAL;

	params->why_failed = 0;

	if (params->format.format > SND_PCM_SFMT_MU_LAW)
		return -EINVAL;

	/* make fragsize even number of samples*channels */
	if (snd_pcm_format_width (params->format.format) > 0)
		sample_size = snd_pcm_format_width (params->format.format) / 8 * params->format.voices;
	else
		return -EINVAL;
	if (params->mode == SND_PCM_MODE_STREAM)
		params->buf.stream.queue_size = (params->buf.stream.queue_size / sample_size) * sample_size;
	else if (params->mode == SND_PCM_MODE_BLOCK)
		params->buf.block.frag_size = (params->buf.block.frag_size / sample_size) * sample_size;
	else
		return -EINVAL;

	/*
	 *  try to decide, if a conversion is required
	 */
	pcm->pre_plugin_format[params->channel] = params->format;
	pcm->start_mode[params->channel] = params->start_mode;
	memcpy (&hwparams, params, sizeof (hwparams));
	pdprintf ("params requested params: format = %i, rate = %i, voices = %i\n",
		hwparams.format.format, hwparams.format.rate, hwparams.format.voices);
	if (ioctl (pcm->fd[params->channel], SND_PCM_IOCTL_CHANNEL_PARAM_FIT, &hwparams) < 0)
	{
		params->why_failed = hwparams.why_failed;
		return -errno;
	}

	pcm->setup_is_valid[params->channel] = 0;
	memset (&pcm->setup[params->channel], 0, sizeof (snd_pcm_channel_setup_t));
	pcm->setup[params->channel].channel = params->channel;
	if ((err = snd_pcm_channel_setup (pcm, &pcm->setup[params->channel])) < 0)
		return err;
	pcm->setup_is_valid[params->channel] = 1;

	snd_pcm_plugin_clear (pcm, params->channel);

	/* add necessary plugins */
	if ((err = snd_pcm_plugin_format (pcm, params, &hwparams)) < 0)
		return err;

	/*
	 *  I/O plugins
	 */

	if (params->mode == SND_PCM_MODE_STREAM)
	{
		pdprintf ("params stream plugin\n");
		err = snd_pcm_plugin_build_stream (pcm, params->channel, &plugin);
	}
	else if (params->mode == SND_PCM_MODE_BLOCK)
	{
		if (!(pcm->plugin_disable_mask & PLUGIN_DISABLE_MMAP) &&
			pcm->setup[params->channel].mmap_valid)
		{
			pdprintf ("params mmap plugin\n");
			err = snd_pcm_plugin_build_mmap (pcm, params->channel, &plugin);
		}
		else
		{
			pdprintf ("params block plugin\n");
			err = snd_pcm_plugin_build_block (pcm, params->channel, &plugin);
		}
	}
	else
	{
		return -EINVAL;
	}
	if (err < 0)
		return err;
	if (params->channel == SND_PCM_CHANNEL_PLAYBACK)
	{
		err = snd_pcm_plugin_append (pcm, params->channel, plugin);
	}
	else
	{
		err = snd_pcm_plugin_insert (pcm, params->channel, plugin);
	}
	if (err < 0)
	{
		snd_pcm_plugin_free (plugin);
		return err;
	}

	if (params->mode == SND_PCM_MODE_BLOCK)
	{
		params->buf.block.frag_size = snd_pcm_plugin_transfer_size (pcm, params->channel,
			pcm->setup[params->channel].buf.block.frag_size);
		pcm->plugin_partial_block_size[params->channel] = params->buf.block.frag_size;
		if (!(pcm->plugin_disable_mask & PLUGIN_DISABLE_BUFFER_PARTIAL_BLOCKS))
		{
			if ((pcm->plugin_partial_block_buffer[params->channel] =
					realloc (pcm->plugin_partial_block_buffer[params->channel],
						pcm->plugin_partial_block_size[params->channel])) == NULL)
				return -ENOMEM;
			pcm->plugin_partial_block_len[params->channel] = 0;
		}
		pcm->plugin_ratio[params->channel] = 1.0 * params->buf.block.frag_size /
			pcm->setup[params->channel].buf.block.frag_size;
		pcm->plugin_cratio[params->channel] = 1.0 * pcm->pre_plugin_format[params->channel].rate /
			pcm->setup[params->channel].format.rate;
	}
	else
	{
		pcm->plugin_partial_block_size[params->channel] = 0;
		pcm->plugin_partial_block_len[params->channel] = 0;
		params->buf.stream.queue_size = snd_pcm_plugin_transfer_size (pcm, params->channel,
			pcm->setup[params->channel].buf.stream.queue_size);
		params->buf.stream.max_fill = params->buf.stream.queue_size;
		pcm->plugin_ratio[params->channel] = 1.0 * params->buf.stream.queue_size /
			pcm->setup[params->channel].buf.stream.queue_size;
		pcm->plugin_cratio[params->channel] = 1.0 * pcm->pre_plugin_format[params->channel].rate /
			pcm->setup[params->channel].format.rate;
	}

	err = snd_pcm_plugin_action (pcm, hwparams.channel, INIT);
	if (err < 0)
		return err;

	return 0;
}

int
snd_pcm_plugin_setup (snd_pcm_t * pcm, snd_pcm_channel_setup_t * setup)
{
	int     err;

	if (!pcm || !setup || setup->channel < 0 || setup->channel > 1)
		return -EINVAL;
	err = snd_pcm_channel_setup (pcm, setup);
	if (err < 0)
		return err;
	if (setup->mode == SND_PCM_MODE_STREAM)
	{
		pdprintf ("params setup: queue_size = %i\n", setup->buf.stream.queue_size);
		setup->buf.stream.queue_size =
			snd_pcm_plugin_transfer_size (pcm, setup->channel, setup->buf.stream.queue_size);
		pdprintf ("params setup: queue_size = %i\n", setup->buf.stream.queue_size);
	}
	else if (setup->mode == SND_PCM_MODE_BLOCK)
	{
		pdprintf ("params setup: frag_size = %i\n", setup->buf.block.frag_size);
		setup->buf.block.frag_size =
			snd_pcm_plugin_transfer_size (pcm, setup->channel, setup->buf.block.frag_size);
		pdprintf ("params setup: frag_size = %i\n", setup->buf.block.frag_size);
	}
	else
	{
		return -EINVAL;
	}
	if (setup->buf.block.max_frag_size == 0)
	{
		if (pcm->plugin_src_mode > SND_SRC_MODE_NORMAL)
			setup->buf.block.max_frag_size = snd_pcm_plugin_src_max_frag(pcm, setup->buf.block.frag_size);
		else
			setup->buf.block.max_frag_size = setup->buf.block.frag_size;
	}

	setup->format = pcm->pre_plugin_format[setup->channel];
	return 0;
}

int
snd_pcm_plugin_status (snd_pcm_t * pcm, snd_pcm_channel_status_t * status)
{
	int     err;

	if (!pcm || !status || status->channel < 0 || status->channel > 1)
		return -EINVAL;
	err = snd_pcm_channel_status (pcm, status);
	if (err < 0)
		return err;
	status->subbuffered = pcm->plugin_partial_block_len[status->channel];
	if (pcm->plugin_src_mode == SND_SRC_MODE_NORMAL)
	{
		status->scount *= pcm->plugin_ratio[status->channel];
		status->count *= pcm->plugin_ratio[status->channel];
		status->free *= pcm->plugin_ratio[status->channel];
	}
	else /* average rate */
	{
		status->scount *= pcm->plugin_cratio[status->channel];
		status->count *= pcm->plugin_cratio[status->channel];
		status->free *= pcm->plugin_cratio[status->channel];
	}
	return 0;
}

int
snd_pcm_plugin_prepare (snd_pcm_t * pcm, int channel)
{
	int     err;

	if ((err = snd_pcm_plugin_action (pcm, channel, PREPARE)) < 0)
		return err;
	return snd_pcm_channel_prepare (pcm, channel);
}

int
snd_pcm_plugin_playback_drain (snd_pcm_t * pcm)
{
	int     err;

	pcm->plugin_partial_block_len[SND_PCM_CHANNEL_PLAYBACK] = 0;
	if ((err = snd_pcm_plugin_action (pcm, SND_PCM_CHANNEL_PLAYBACK, DRAIN)) < 0)
		return err;
	return snd_pcm_playback_drain (pcm);
}

int
snd_pcm_plugin_flush (snd_pcm_t * pcm, int channel)
{
	int     n = 0;
	unsigned char silence;
	int     err;
	int     rtn;

	pdprintf ("flush\n");
	if (channel == SND_PCM_CHANNEL_PLAYBACK && pcm->plugin_partial_block_len[channel] != 0)
	{
		if (pcm->setup_is_valid[SND_PCM_CHANNEL_PLAYBACK])
			silence = snd_pcm_plugin_silence (&pcm->pre_plugin_format[SND_PCM_CHANNEL_PLAYBACK]);
		else
			silence = 0x0;
		n = pcm->plugin_partial_block_size[SND_PCM_CHANNEL_PLAYBACK] -
			pcm->plugin_partial_block_len[SND_PCM_CHANNEL_PLAYBACK];
		memset (pcm->plugin_partial_block_buffer[SND_PCM_CHANNEL_PLAYBACK] +
			pcm->plugin_partial_block_len[SND_PCM_CHANNEL_PLAYBACK], silence, n);
		snd_pcm_plugin_write1 (pcm, pcm->plugin_partial_block_buffer[channel],
			pcm->plugin_partial_block_size[channel]);
	}
	pcm->plugin_partial_block_len[channel] = 0;
	if ((err = snd_pcm_plugin_action (pcm, channel, FLUSH)) < 0)
		return err;
	if ((rtn = snd_pcm_channel_flush (pcm, channel)) != 0)
		return rtn;
	else
		return n;
}

int
snd_pcm_plugin_pointer (snd_pcm_t * pcm, int channel, void **ptr, size_t * size)
{
	snd_pcm_plugin_t *plugin;
	int     err;

	if (!ptr || !size)
		return -EINVAL;
	*ptr = NULL;
	if (!pcm || channel < 0 || channel > 1)
		return -EINVAL;
	plugin = pcm->plugin_first[channel];
	if (!plugin)
		return -EINVAL;
	if (plugin->transfer_src_ptr)
	{
		err = plugin->transfer_src_ptr (plugin, ptr, size);
		if (err >= 0)
			return 0;
	}
	if (pcm->plugin_alloc_xptr[channel])
	{
		if (pcm->plugin_alloc_xsize[channel] >= *size)
		{
			*ptr = (char *) pcm->plugin_alloc_xptr[channel];
			return 0;
		}
		*ptr = (char *) realloc (pcm->plugin_alloc_xptr[channel], *size);
	}
	else
	{
		*ptr = (char *) malloc (*size);
		if (*ptr != NULL)
			pcm->plugin_alloc_xsize[channel] = *size;
	}
	if (*ptr == NULL)
		return -ENOMEM;
	pcm->plugin_alloc_xptr[channel] = *ptr;
	return 0;
}

static void *
snd_pcm_plugin_malloc (snd_pcm_t * pcm, long size)
{
	int     idx;
	void   *ptr;

	for (idx = 0; idx < 4; idx++)
	{
		if (pcm->plugin_alloc_lock[idx])
			continue;
		if (pcm->plugin_alloc_ptr[idx] == NULL)
			continue;
		if (pcm->plugin_alloc_size[idx] >= size)
		{
			pcm->plugin_alloc_lock[idx] = 1;
			return pcm->plugin_alloc_ptr[idx];
		}
	}
	for (idx = 0; idx < 4; idx++)
	{
		if (pcm->plugin_alloc_lock[idx])
			continue;
		if (pcm->plugin_alloc_ptr[idx] == NULL)
			continue;
		ptr = realloc (pcm->plugin_alloc_ptr[idx], size);
		if (ptr == NULL)
			continue;
		pcm->plugin_alloc_size[idx] = size;
		pcm->plugin_alloc_lock[idx] = 1;
		return pcm->plugin_alloc_ptr[idx] = ptr;
	}
	for (idx = 0; idx < 4; idx++)
	{
		if (pcm->plugin_alloc_ptr[idx] != NULL)
			continue;
		ptr = malloc (size);
		if (ptr == NULL)
			continue;
		pcm->plugin_alloc_size[idx] = size;
		pcm->plugin_alloc_lock[idx] = 1;
		return pcm->plugin_alloc_ptr[idx] = ptr;
	}
	return NULL;
}

static int
snd_pcm_plugin_alloc_unlock (snd_pcm_t * pcm, void *ptr)
{
	int     idx;

	for (idx = 0; idx < 4; idx++)
	{
		if (pcm->plugin_alloc_ptr[idx] == ptr)
		{
			pcm->plugin_alloc_lock[idx] = 0;
			return 0;
		}
	}
	return -ENOENT;
}

static  ssize_t
snd_pcm_plugin_write1 (snd_pcm_t * pcm, const void *buffer, size_t count)
{
	snd_pcm_plugin_t *plugin, *next;
	void   *dst_ptr, *dst_ptr1 = NULL, *src_ptr, *src_ptr1 = NULL;
	size_t  dst_size, src_size;
	ssize_t size = 0, result = 0;
	int     err;
	snd_pcm_channel_setup_t setup;

	memset (&setup, 0, sizeof (setup));
	setup.channel = SND_PCM_CHANNEL_PLAYBACK;
	if ((err = snd_pcm_plugin_setup (pcm, &setup)) < 0)
		return (err);
	if (setup.mode == SND_PCM_MODE_BLOCK && pcm->mode & SND_PCM_OPEN_NONBLOCK)
	{
		int     fdsetsize;
		fd_set *wfdset;
		struct timeval tv;
		fdsetsize =
			(pcm->fd[SND_PCM_CHANNEL_PLAYBACK] + sizeof (int) * CHAR_BIT -
			1) / (sizeof (int) * CHAR_BIT) * sizeof (int);
		wfdset = (struct fd_set *) alloca (fdsetsize);
		memset (wfdset, 0, fdsetsize);
		FD_SET (pcm->fd[SND_PCM_CHANNEL_PLAYBACK], wfdset);
		tv.tv_sec = 0;
		tv.tv_usec = 0;
		switch (result = select (1 + pcm->fd[SND_PCM_CHANNEL_PLAYBACK], 0, wfdset, 0, &tv))
		{
		case -1:
			return -EIO;
		case 0:
			return -EWOULDBLOCK;
		}
	}

	if ((plugin = snd_pcm_plugin_first (pcm, SND_PCM_CHANNEL_PLAYBACK)) == NULL)
		return snd_pcm_write (pcm, buffer, count);
	src_ptr = (char *) buffer;
	dst_size = src_size = count;
	while (plugin)
	{
		next = plugin->next;
		if (plugin->dst_size)
		{
			dst_size = plugin->dst_size (plugin, dst_size);
			if (dst_size < 0)
			{
				result = dst_size;
				goto __free;
			}
		}
		if (next != NULL)
		{
			if (next->transfer_src_ptr)
			{
				if ((err = next->transfer_src_ptr (next, &dst_ptr, &dst_size)) < 0)
				{
					if (dst_ptr == NULL)
						goto __alloc;
					result = err;
					goto __free;
				}
			}
			else
			{
			  __alloc:
				dst_ptr = dst_ptr1 = (char *) snd_pcm_plugin_malloc (pcm, dst_size);
				if (dst_ptr == NULL)
				{
					result = -ENOMEM;
					goto __free;
				}
			}
		}
		else
		{
			dst_ptr = src_ptr;
			dst_size = src_size;
		}
		pdprintf ("write plugin: %s, %i, %i\n", plugin->name, src_size, dst_size);
		if ((size = plugin->transfer (plugin, src_ptr, src_size, dst_ptr, dst_size)) < 0)
		{
			result = size;
			goto __free;
		}
		if (src_ptr1)
			snd_pcm_plugin_alloc_unlock (pcm, src_ptr1);
		plugin = next;
		src_ptr = dst_ptr;
		src_ptr1 = dst_ptr1;
		dst_ptr1 = NULL;
		src_size = dst_size = size;
	}
	result = snd_pcm_plugin_transfer_size (pcm, SND_PCM_CHANNEL_PLAYBACK, size);
	pdprintf ("size = %i, result = %i, count = %i\n", size, result, count);
  __free:
	if (dst_ptr1)
		snd_pcm_plugin_alloc_unlock (pcm, dst_ptr1);
	if (src_ptr1)
		snd_pcm_plugin_alloc_unlock (pcm, src_ptr1);
	return result;
}

ssize_t
snd_pcm_plugin_write (snd_pcm_t * pcm, const void *buffer, size_t count)
{
	char   *ptr;
	size_t  n;
	ssize_t total, rtn;

	/* if !buffering & block_mode & count!=block_size */

	/* BLOCK_MODE */
	if (pcm->plugin_partial_block_size[SND_PCM_CHANNEL_PLAYBACK] != 0 && pcm->plugin_src_mode == SND_SRC_MODE_NORMAL)
	{
		if ((pcm->plugin_disable_mask & PLUGIN_DISABLE_BUFFER_PARTIAL_BLOCKS) &&
			pcm->plugin_partial_block_size[SND_PCM_CHANNEL_PLAYBACK] != count)
			return -EINVAL;
		else
		{
			for (ptr = (char *) buffer, total = 0; count > 0;)
			{											   /* do we have a partial block subbuffered | less then a full block */
				if (pcm->plugin_partial_block_len[SND_PCM_CHANNEL_PLAYBACK] > 0 ||
					count < pcm->plugin_partial_block_size[SND_PCM_CHANNEL_PLAYBACK])
				{										   /* if we can't make a full block just add to subbuffer */
					if (count <
						(pcm->plugin_partial_block_size[SND_PCM_CHANNEL_PLAYBACK] -
							pcm->plugin_partial_block_len[SND_PCM_CHANNEL_PLAYBACK]))
					{
						memcpy (pcm->plugin_partial_block_buffer[SND_PCM_CHANNEL_PLAYBACK] +
							pcm->plugin_partial_block_len[SND_PCM_CHANNEL_PLAYBACK], ptr, count);
						pcm->plugin_partial_block_len[SND_PCM_CHANNEL_PLAYBACK] += count;
						total += count;
						count = 0;
					}
					else
					{									   /* make a full block and write it out */
						n = pcm->plugin_partial_block_size[SND_PCM_CHANNEL_PLAYBACK] -
							pcm->plugin_partial_block_len[SND_PCM_CHANNEL_PLAYBACK];
						memcpy (pcm->plugin_partial_block_buffer[SND_PCM_CHANNEL_PLAYBACK] +
							pcm->plugin_partial_block_len[SND_PCM_CHANNEL_PLAYBACK], ptr, n);
						if ((rtn =
								snd_pcm_plugin_write1 (pcm,
									pcm->plugin_partial_block_buffer[SND_PCM_CHANNEL_PLAYBACK],
									pcm->plugin_partial_block_size[SND_PCM_CHANNEL_PLAYBACK])) < 0)
						{
							errno = -rtn;
							break;						   /* error */
						}
						pcm->plugin_partial_block_len[SND_PCM_CHANNEL_PLAYBACK] = 0;
						total += n;
						count -= n;
						ptr += n;
					}
				}
				else
				{										   /* write a full block directly */
					if ((rtn =
							snd_pcm_plugin_write1 (pcm, ptr,
								pcm->plugin_partial_block_size[SND_PCM_CHANNEL_PLAYBACK])) < 0)
					{
						errno = -rtn;
						break;							   /* error */
					}
					total += pcm->plugin_partial_block_size[SND_PCM_CHANNEL_PLAYBACK];
					count -= pcm->plugin_partial_block_size[SND_PCM_CHANNEL_PLAYBACK];
					ptr += pcm->plugin_partial_block_size[SND_PCM_CHANNEL_PLAYBACK];
				}
			}
		}
	}
	else
		total = snd_pcm_plugin_write1 (pcm, buffer, count);
	return (total);
}

static  ssize_t
snd_pcm_plugin_read1 (snd_pcm_t * pcm, void *buffer, size_t count)
{
	snd_pcm_plugin_t *plugin, *next;
	void   *dst_ptr, *dst_ptr1 = NULL, *src_ptr, *src_ptr1 = NULL;
	size_t  dst_size, src_size;
	ssize_t size = 0, result = 0;
	int     err;

	if ((plugin = snd_pcm_plugin_first (pcm, SND_PCM_CHANNEL_CAPTURE)) == NULL)
		return snd_pcm_read (pcm, buffer, count);
	src_ptr = NULL;
	src_size = 0;
	dst_size = snd_pcm_plugin_hardware_size (pcm, SND_PCM_CHANNEL_CAPTURE, count);
	if (dst_size < 0)
		return dst_size;
	while (plugin)
	{
		next = plugin->next;
		if (plugin->dst_size)
		{
			dst_size = plugin->dst_size (plugin, dst_size);
			if (dst_size < 0)
			{
				result = dst_size;
				goto __free;
			}
		}
		if (next != NULL)
		{
			if (next->transfer_src_ptr)
			{
				if ((err = next->transfer_src_ptr (next, &dst_ptr, &dst_size)) < 0)
				{
					if (dst_ptr == NULL)
						goto __alloc;
					result = err;
					goto __free;
				}
			}
			else
			{
			  __alloc:
				dst_ptr = dst_ptr1 = (char *) snd_pcm_plugin_malloc (pcm, dst_size);
				if (dst_ptr == NULL)
				{
					result = -ENOMEM;
					goto __free;
				}
			}
		}
		else
		{
			dst_ptr = buffer;
		}
		pdprintf ("read plugin: %s, %i, %i\n", plugin->name, src_size, dst_size);
		if ((size = plugin->transfer (plugin, src_ptr, src_size, dst_ptr, dst_size)) < 0)
		{
			result = size;
			goto __free;
		}
		if (src_ptr1)
			snd_pcm_plugin_alloc_unlock (pcm, src_ptr1);
		plugin = plugin->next;
		src_ptr = dst_ptr;
		src_ptr1 = dst_ptr1;
		dst_ptr1 = NULL;
		src_size = dst_size = size;
	}
	result = size;
  __free:
	if (dst_ptr1)
		snd_pcm_plugin_alloc_unlock (pcm, dst_ptr1);
	if (src_ptr1)
		snd_pcm_plugin_alloc_unlock (pcm, src_ptr1);
	return result;
}

ssize_t
snd_pcm_plugin_read (snd_pcm_t * pcm, void *buffer, size_t count)
{
	char   *ptr = NULL;
	ssize_t total = 0, rtn = 0;

	/* BLOCK_MODE */
	if (pcm->plugin_partial_block_size[SND_PCM_CHANNEL_CAPTURE] != 0)
	{
		if ((pcm->plugin_disable_mask & PLUGIN_DISABLE_BUFFER_PARTIAL_BLOCKS) &&
			pcm->plugin_partial_block_size[SND_PCM_CHANNEL_CAPTURE] != count)
			return -EINVAL;
		else
		{
			for (ptr = (char *) buffer, total = 0; count > 0;)
			{											   /* do we have a partial block remaining from a previous read? */
				if (pcm->plugin_partial_block_len[SND_PCM_CHANNEL_CAPTURE])
				{										   /* do we have enough already in the partial_block_buffer? */
					if (count <= pcm->plugin_partial_block_len[SND_PCM_CHANNEL_CAPTURE])
					{
						memcpy (ptr, pcm->plugin_partial_block_buffer[SND_PCM_CHANNEL_CAPTURE],
							count);
						pcm->plugin_partial_block_len[SND_PCM_CHANNEL_CAPTURE] -= count;
						memcpy (pcm->plugin_partial_block_buffer[SND_PCM_CHANNEL_CAPTURE],
							pcm->plugin_partial_block_buffer[SND_PCM_CHANNEL_CAPTURE] + count,
							pcm->plugin_partial_block_len[SND_PCM_CHANNEL_CAPTURE]);
						ptr += count;
						total += count;
						count = 0;
					}
					else
					{
						memcpy (ptr, pcm->plugin_partial_block_buffer[SND_PCM_CHANNEL_CAPTURE],
							pcm->plugin_partial_block_len[SND_PCM_CHANNEL_CAPTURE]);
						ptr += pcm->plugin_partial_block_len[SND_PCM_CHANNEL_CAPTURE];
						total += pcm->plugin_partial_block_len[SND_PCM_CHANNEL_CAPTURE];
						count -= pcm->plugin_partial_block_len[SND_PCM_CHANNEL_CAPTURE];
						pcm->plugin_partial_block_len[SND_PCM_CHANNEL_CAPTURE] = 0;
					}

					memset (pcm->plugin_partial_block_buffer[SND_PCM_CHANNEL_CAPTURE] +
						pcm->plugin_partial_block_len[SND_PCM_CHANNEL_CAPTURE], 0,
						pcm->plugin_partial_block_size[SND_PCM_CHANNEL_CAPTURE] -
						pcm->plugin_partial_block_len[SND_PCM_CHANNEL_CAPTURE]);
				}
				/* do we want less than one real block? */
				else if (count < pcm->plugin_partial_block_size[SND_PCM_CHANNEL_CAPTURE])
				{										   /* read in a full block */
					if ((rtn =
							snd_pcm_plugin_read1 (pcm,
								pcm->plugin_partial_block_buffer[SND_PCM_CHANNEL_CAPTURE],
								pcm->plugin_partial_block_size[SND_PCM_CHANNEL_CAPTURE])) < 0)
					{
						errno = -rtn;
						pdprintf
							("snd_pcm_plugin_read error: unable to read partial block - %s (%i)\n",
							snd_strerror (rtn), rtn);
						break;							   /* error */
					}
					memcpy (ptr, pcm->plugin_partial_block_buffer[SND_PCM_CHANNEL_CAPTURE], count);
					pcm->plugin_partial_block_len[SND_PCM_CHANNEL_CAPTURE] =
						pcm->plugin_partial_block_size[SND_PCM_CHANNEL_CAPTURE] - count;
					memcpy (pcm->plugin_partial_block_buffer[SND_PCM_CHANNEL_CAPTURE],
						pcm->plugin_partial_block_buffer[SND_PCM_CHANNEL_CAPTURE] + count,
						pcm->plugin_partial_block_len[SND_PCM_CHANNEL_CAPTURE]);
					ptr += count;
					total += count;
					count = 0;
				}
				else
				{										   /* read a full block directly */
					if ((rtn =
							snd_pcm_plugin_read1 (pcm, ptr,
								pcm->plugin_partial_block_size[SND_PCM_CHANNEL_CAPTURE])) < 0)
					{
						errno = -rtn;
						pdprintf
							("snd_pcm_plugin_read error: unable to read full block - %s (%i)\n",
							snd_strerror (rtn), rtn);
						break;							   /* error */
					}
					ptr += pcm->plugin_partial_block_size[SND_PCM_CHANNEL_CAPTURE];
					total += pcm->plugin_partial_block_size[SND_PCM_CHANNEL_CAPTURE];
					count -= pcm->plugin_partial_block_size[SND_PCM_CHANNEL_CAPTURE];
				}
			}
		}
	}
	else
		total = snd_pcm_plugin_read1 (pcm, buffer, count);

	return (total);
}
