/*
 *  Rate conversion Plug-In
 *  Copyright (c) 1999 by Jaroslav Kysela <perex@suse.cz>
 *
 *
 *   This library is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Library General Public License as
 *   published by the Free Software Foundation; either version 2 of
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU Library General Public License for more details.
 *
 *   You should have received a copy of the GNU Library General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <endian.h>
#include <byteswap.h>
#include "../pcm_local.h"
#include <sys/slog.h>
 #include <sys/slogcodes.h>

__SRCVERSION( "$URL: http://svn.ott.qnx.com/product/branches/6.5.0/SP1/lib/asound/pcm/plugin/rate.c $ $Rev: 646317 $" );


#define SHIFT	30
#define BITS	(1<<SHIFT)
#define MASK	(BITS-1)
#define MAX_VOICES 8
#define CNTRESET 1000000.0f
#define MAXPHASE asrc_maxphase
#define MINPHASE asrc_minphase
#define ASRCTC   asrc_tc
#define ASRCADJ  asrc_adj
#define ASRCDAMP asrc_damp
#define ASRCDEBUG asrc_debug

float asrc_maxphase = 1.0025f; //1.02f //1.01f
float asrc_minphase = 0.9975f; //0.98f //0.99f
float asrc_tc = 0.9f; //0.6f //0.999
int asrc_adj = 1; //2
int asrc_damp = 1;
int asrc_debug = 0;

static int asrc_conf_done = 0;


/*
 *  Basic rate conversion plugin
 */

struct rate_private_data;
typedef void (*resample_f) (struct rate_private_data * data, int voices, char *src_ptr,
	int src_size, char *dst_ptr, int dst_size);
typedef signed short (*take_sample_f) (void *ptr);
typedef void (*put_sample_f) (void *ptr, signed int val);

struct rate_private_data
{
	snd_pcm_format_t format;
	int     src_voices;
	int     dst_voices;
	int     src_rate;
	int     dst_rate;
	int     sample_size;
	resample_f method;
	take_sample_f take;
	put_sample_f put;
	int64_t pitch;
	ssize_t old_src_size, old_dst_size;

	unsigned int src_mode;
	int	icount;
	int ocount;
	int newfragsize;
	double src_ratio;
	double fractfrag;
	int target;
	double filttarget;
	int src_offset;
	int maxfragsize;
	int minfragsize;
	int dstfragsize;
	int initialized;

	/* used for the linear converter */
	int16_t S[MAX_VOICES];

	/* used for the band limited converter */
	uint16_t Nmult;						/* Filter length for up-conversions */
	int16_t *Imp;						/* Filter coefficients */
	int16_t *ImpD;						/* ImpD[n] = Imp[n+1]-Imp[n] */
	uint16_t LpScl;						/* Unity-gain scale factor */
	uint16_t Nwing;						/* Filter table size */
	int16_t *X[MAX_VOICES];
};


#define RATE_TAKE_SAMPLE(name, type, val) \
static signed short rate_take_sample_##name(void *ptr) \
{ \
	signed int smp = *(type *)ptr; \
	return val; \
}

#define RATE_PUT_SAMPLE(name, type, val) \
static void rate_put_sample_##name(void *ptr, signed int smp) \
{ \
	*(type *)ptr = val; \
}

#if __BYTE_ORDER == __LITTLE_ENDIAN
#define my_little_swap16(x) (x)
#define my_little_swap32(x) (x)
#define my_big_swap16(x) bswap_16(x)
#define my_big_swap32(x) bswap_32(x)
#else
#define my_little_swap16(x) bswap_16(x)
#define my_little_swap32(x) bswap_32(x)
#define my_big_swap16(x) (x)
#define my_big_swap32(x) (x)
#endif

RATE_TAKE_SAMPLE (s8, int8_t, smp << 8)
RATE_TAKE_SAMPLE (u8, int8_t, (smp << 8) ^ 0x8000)
RATE_TAKE_SAMPLE (s16_le, int16_t, my_little_swap16 (smp))
RATE_TAKE_SAMPLE (s16_be, int16_t, my_big_swap16 (smp))
RATE_TAKE_SAMPLE (u16_le, int16_t, my_little_swap16 (smp) ^ 0x8000)
RATE_TAKE_SAMPLE (u16_be, int16_t, my_big_swap16 (smp) ^ 0x8000)
RATE_TAKE_SAMPLE (s24_le, int32_t, my_little_swap32 (smp) >> 8)
RATE_TAKE_SAMPLE (s24_be, int32_t, my_big_swap32 (smp) >> 8)
RATE_TAKE_SAMPLE (u24_le, int32_t, (my_little_swap32 (smp) >> 8) ^ 0x8000)
RATE_TAKE_SAMPLE (u24_be, int32_t, (my_big_swap32 (smp) >> 8) ^ 0x8000)
RATE_TAKE_SAMPLE (s32_le, int32_t, my_little_swap32 (smp) >> 16)
RATE_TAKE_SAMPLE (s32_be, int32_t, my_big_swap32 (smp) >> 16)
RATE_TAKE_SAMPLE (u32_le, int32_t, (my_little_swap32 (smp) >> 16) ^ 0x8000)
RATE_TAKE_SAMPLE (u32_be, int32_t, (my_big_swap32 (smp) >> 16) ^ 0x8000)
	 static take_sample_f rate_take_sample[] = {
		 [SND_PCM_SFMT_S8] rate_take_sample_s8,
		 [SND_PCM_SFMT_U8] rate_take_sample_u8,
		 [SND_PCM_SFMT_S16_LE] rate_take_sample_s16_le,
		 [SND_PCM_SFMT_S16_BE] rate_take_sample_s16_be,
		 [SND_PCM_SFMT_U16_LE] rate_take_sample_u16_le,
		 [SND_PCM_SFMT_U16_BE] rate_take_sample_u16_be,
		 [SND_PCM_SFMT_S24_LE] rate_take_sample_s24_le,
		 [SND_PCM_SFMT_S24_BE] rate_take_sample_s24_be,
		 [SND_PCM_SFMT_U24_LE] rate_take_sample_u24_le,
		 [SND_PCM_SFMT_U24_BE] rate_take_sample_u24_be,
		 [SND_PCM_SFMT_S32_LE] rate_take_sample_s32_le,
		 [SND_PCM_SFMT_S32_BE] rate_take_sample_s32_be,
		 [SND_PCM_SFMT_U32_LE] rate_take_sample_u32_le,
		 [SND_PCM_SFMT_U32_BE] rate_take_sample_u32_be
	 };

RATE_PUT_SAMPLE (s8, int8_t, smp >> 8)
RATE_PUT_SAMPLE (u8, int8_t, (smp >> 8) ^ 0x80)
RATE_PUT_SAMPLE (s16_le, int16_t, my_little_swap16 (smp))
RATE_PUT_SAMPLE (s16_be, int16_t, my_big_swap16 (smp))
RATE_PUT_SAMPLE (u16_le, int16_t, my_little_swap16 (smp ^ 0x8000))
RATE_PUT_SAMPLE (u16_be, int16_t, my_big_swap16 (smp ^ 0x8000))
RATE_PUT_SAMPLE (s24_le, int32_t, my_little_swap32 (smp << 8))
RATE_PUT_SAMPLE (s24_be, int32_t, my_big_swap32 (smp << 8))
RATE_PUT_SAMPLE (u24_le, int32_t, my_little_swap32 ((smp ^ 0x8000) << 8))
RATE_PUT_SAMPLE (u24_be, int32_t, my_big_swap32 ((smp ^ 0x8000) << 8))
RATE_PUT_SAMPLE (s32_le, int32_t, my_little_swap32 (smp << 16))
RATE_PUT_SAMPLE (s32_be, int32_t, my_big_swap32 (smp << 16))
RATE_PUT_SAMPLE (u32_le, int32_t, my_little_swap32 ((smp ^ 0x8000) << 16))
RATE_PUT_SAMPLE (u32_be, int32_t, my_big_swap32 ((smp ^ 0x8000) << 16))
	 static put_sample_f rate_put_sample[] = {
		 [SND_PCM_SFMT_S8] rate_put_sample_s8,
		 [SND_PCM_SFMT_U8] rate_put_sample_u8,
		 [SND_PCM_SFMT_S16_LE] rate_put_sample_s16_le,
		 [SND_PCM_SFMT_S16_BE] rate_put_sample_s16_be,
		 [SND_PCM_SFMT_U16_LE] rate_put_sample_u16_le,
		 [SND_PCM_SFMT_U16_BE] rate_put_sample_u16_be,
		 [SND_PCM_SFMT_S24_LE] rate_put_sample_s24_le,
		 [SND_PCM_SFMT_S24_BE] rate_put_sample_s24_be,
		 [SND_PCM_SFMT_U24_LE] rate_put_sample_u24_le,
		 [SND_PCM_SFMT_U24_BE] rate_put_sample_u24_be,
		 [SND_PCM_SFMT_S32_LE] rate_put_sample_s32_le,
		 [SND_PCM_SFMT_S32_BE] rate_put_sample_s32_be,
		 [SND_PCM_SFMT_U32_LE] rate_put_sample_u32_le,
		 [SND_PCM_SFMT_U32_BE] rate_put_sample_u32_be
	 };

/******************************************************************************/
/* BSD License:                                                               */
/*                                                                            */
/* Copyright (c) 2003.  Julius O. Smith III . All Rights Reserved.            */
/*                                                                            */
/* Redistribution and use in source and binary forms, with or without         */
/* modifications, are permitted provided that the following conditions are    */
/* met:                                                                       */
/* i.          Redistribution of source code must retain the above copyright  */
/* notice, this list of conditions and the following disclaimer.              */
/* ii.         Redistribution in binary form must reproduce the above         */
/* copyright notice, this list of conditions and the following disclaimer in  */
/* the documentation and/or other materials provided with the distribution.   */
/* iii.        Neither the name(s) of Julius O. Smith III  nor the names of   */
/* its/their contributors may be used to endorse or promote products derived  */
/* from this software without specific prior written permission.              */
/*                                                                            */
/* THIS SOFTWARE IS PROVIDED BY Julius O. Smith III  AND CONTRIBUTORS         */
/* "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING , BUT NOT LIMITED */
/* TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR */
/* PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE [NAME OF LICENSOR(S)] OR    */
/* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,      */
/* EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,        */
/* PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR         */
/* PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF     */
/* LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING       */
/* NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS         */
/* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.               */
/******************************************************************************/
/* Conversion constants */
#define Nhc       8
#define Na        7
#define Np       (Nhc+Na)
#define Npc      (1<<Nhc)
#define Amask    ((1<<Na)-1)
#define Pmask    ((1<<Np)-1)
#define Nh       16
#define Nb       16
#define Nhxn     14
#define Nhg      (Nh-Nhxn)
#define NLpScl   13

#define	USE_INTERP		1


/* *INDENT-OFF* */
#define A_FILTER_NMULT 7
#define A_FILTER_SCALE 13175
#define A_FILTER_NWING 768
static int16_t A_FILTER_IMP[] = {
 32767, 32766, 32764, 32760, 32755, 32748, 32740, 32730, 32718, 32705, 32691, 32675, 32657, 32638,
 32618, 32596, 32572, 32547, 32521, 32493, 32463, 32432, 32400, 32366, 32330, 32293, 32255, 32215,
 32173, 32131, 32086, 32041, 31993, 31945, 31895, 31843, 31790, 31736, 31680, 31623, 31565, 31505,
 31444, 31381, 31317, 31251, 31185, 31117, 31047, 30976, 30904, 30831, 30756, 30680, 30602, 30524,
 30444, 30363, 30280, 30197, 30112, 30025, 29938, 29849, 29760, 29669, 29576, 29483, 29388, 29293,
 29196, 29098, 28999, 28899, 28797, 28695, 28591, 28487, 28381, 28275, 28167, 28058, 27948, 27838,
 27726, 27613, 27499, 27385, 27269, 27153, 27035, 26917, 26798, 26678, 26557, 26435, 26312, 26189,
 26064, 25939, 25813, 25686, 25559, 25431, 25302, 25172, 25042, 24910, 24779, 24646, 24513, 24379,
 24245, 24110, 23974, 23838, 23701, 23564, 23426, 23287, 23148, 23008, 22868, 22728, 22587, 22445,
 22303, 22161, 22018, 21875, 21732, 21588, 21443, 21299, 21154, 21008, 20863, 20717, 20571, 20424,
 20278, 20131, 19983, 19836, 19688, 19541, 19393, 19245, 19096, 18948, 18799, 18651, 18502, 18353,
 18204, 18055, 17907, 17758, 17609, 17460, 17311, 17162, 17013, 16864, 16715, 16566, 16418, 16269,
 16121, 15972, 15824, 15676, 15528, 15381, 15233, 15086, 14939, 14792, 14645, 14499, 14353, 14207,
 14061, 13916, 13771, 13626, 13482, 13338, 13194, 13050, 12907, 12765, 12623, 12481, 12339, 12198,
 12058, 11918, 11778, 11639, 11500, 11362, 11224, 11086, 10950, 10813, 10677, 10542, 10407, 10273,
 10140, 10007,  9874,  9742,  9611,  9480,  9350,  9221,  9092,  8963,  8836,  8709,  8582,  8457,
  8332,  8207,  8084,  7961,  7838,  7717,  7596,  7476,  7356,  7237,  7119,  7002,  6886,  6770,
  6655,  6540,  6427,  6314,  6202,  6090,  5980,  5870,  5761,  5653,  5546,  5439,  5333,  5228,
  5124,  5021,  4918,  4816,  4715,  4615,  4516,  4417,  4320,  4223,  4127,  4032,  3938,  3844,
  3751,  3660,  3569,  3479,  3389,  3301,  3213,  3127,  3041,  2956,  2872,  2788,  2706,  2624,
  2543,  2463,  2384,  2306,  2229,  2152,  2077,  2002,  1928,  1855,  1782,  1711,  1640,  1571,
  1502,  1434,  1367,  1300,  1235,  1170,  1106,  1043,   981,   919,   859,   799,   740,   682,
   625,   568,   513,   458,   404,   351,   298,   246,   196,   146,    96,    48,     0,   -46,
   -92,  -138,  -182,  -226,  -269,  -312,  -353,  -394,  -434,  -474,  -513,  -551,  -588,  -624,
  -660,  -695,  -730,  -763,  -797,  -829,  -861,  -891,  -922,  -951,  -980, -1009, -1036, -1063,
 -1090, -1115, -1141, -1165, -1189, -1212, -1235, -1257, -1278, -1299, -1319, -1339, -1358, -1376,
 -1394, -1412, -1428, -1445, -1460, -1475, -1490, -1504, -1518, -1531, -1543, -1555, -1567, -1578,
 -1588, -1599, -1608, -1617, -1626, -1634, -1642, -1649, -1656, -1662, -1668, -1674, -1679, -1684,
 -1688, -1692, -1696, -1699, -1701, -1704, -1706, -1707, -1709, -1710, -1710, -1710, -1710, -1710,
 -1709, -1708, -1706, -1705, -1703, -1700, -1697, -1695, -1691, -1688, -1684, -1680, -1676, -1671,
 -1666, -1661, -1656, -1650, -1645, -1638, -1632, -1626, -1619, -1612, -1605, -1598, -1590, -1583,
 -1575, -1567, -1559, -1550, -1542, -1533, -1524, -1515, -1506, -1497, -1488, -1478, -1468, -1458,
 -1449, -1439, -1428, -1418, -1408, -1397, -1387, -1376, -1365, -1355, -1344, -1333, -1322, -1311,
 -1299, -1288, -1277, -1265, -1254, -1243, -1231, -1220, -1208, -1196, -1185, -1173, -1161, -1150,
 -1138, -1126, -1114, -1103, -1091, -1079, -1067, -1055, -1044, -1032, -1020, -1008,  -997,  -985,
  -973,  -961,  -950,  -938,  -926,  -915,  -903,  -892,  -880,  -869,  -857,  -846,  -835,  -823,
  -812,  -801,  -790,  -779,  -768,  -757,  -746,  -735,  -724,  -714,  -703,  -692,  -682,  -671,
  -661,  -650,  -640,  -630,  -620,  -610,  -600,  -590,  -580,  -570,  -561,  -551,  -542,  -532,
  -523,  -514,  -504,  -495,  -486,  -477,  -469,  -460,  -451,  -442,  -434,  -426,  -417,  -409,
  -401,  -393,  -385,  -377,  -369,  -361,  -354,  -346,  -339,  -332,  -324,  -317,  -310,  -303,
  -296,  -289,  -283,  -276,  -269,  -263,  -257,  -250,  -244,  -238,  -232,  -226,  -220,  -215,
  -209,  -203,  -198,  -192,  -187,  -182,  -177,  -172,  -167,  -162,  -157,  -152,  -147,  -143,
  -138,  -134,  -130,  -125,  -121,  -117,  -113,  -109,  -105,  -101,   -98,   -94,   -91,   -87,
   -84,   -80,   -77,   -74,   -71,   -67,   -64,   -61,   -59,   -56,   -53,   -50,   -48,   -45,
   -43,   -40,   -38,   -35,   -33,   -31,   -29,   -27,   -25,   -23,   -21,   -19,   -17,   -15,
   -14,   -12,   -10,    -9,    -7,    -6,    -4,    -3,    -2,     0,     0,     1,     2,     4,
     5,     6,     7,     8,     9,     9,    10,    11,    12,    13,    13,    14,    15,    15,
    16,    16,    17,    18,    18,    18,    19,    19,    20,    20,    20,    21,    21,    21,
    21,    21,    22,    22,    22,    22,    22,    22,    22,    22,    22,    22,    22,    22,
    22,    22,    22,    22,    22,    22,    22,    22,    22,    22,    21,    21,    21,    21,
    21,    21,    20,    20,    20,    20,    20,    19,    19,    19,    19,    18,    18,    18,
    18,    17,    17,    17,    16,    16,    16,    16,    15,    15,    15,    14,    14,    14,
    14,    13,    13,    13,    12,    12,    12,    12,    11,    11,    11,    10,    10,    10,
    10,     9,     9,     9,     9,     8,     8,     8,     8,     7,     7,     7,     7,     6,
     6,     6,     6,     5,     5,     5,     5,     5,     4,     4,     4,     4,};
static int16_t A_FILTER_IMPD[] = {
  -1,  -2,  -4,  -5,  -7,  -8, -10, -12, -13, -14, -16, -18, -19, -20, -22, -24, -25, -26, -28, -30,
 -31, -32, -34, -36, -37, -38, -40, -42, -42, -45, -45, -48, -48, -50, -52, -53, -54, -56, -57, -58,
 -60, -61, -63, -64, -66, -66, -68, -70, -71, -72, -73, -75, -76, -78, -78, -80, -81, -83, -83, -85,
 -87, -87, -89, -89, -91, -93, -93, -95, -95, -97, -98, -99,-100,-102,-102,-104,-104,-106,-106,-108,
-109,-110,-110,-112,-113,-114,-114,-116,-116,-118,-118,-119,-120,-121,-122,-123,-123,-125,-125,-126,
-127,-127,-128,-129,-130,-130,-132,-131,-133,-133,-134,-134,-135,-136,-136,-137,-137,-138,-139,-139,
-140,-140,-140,-141,-142,-142,-142,-143,-143,-143,-144,-145,-144,-145,-146,-145,-146,-146,-147,-146,
-147,-148,-147,-148,-147,-148,-148,-149,-148,-149,-148,-149,-149,-149,-149,-148,-149,-149,-149,-149,
-149,-149,-149,-149,-149,-148,-149,-148,-149,-148,-148,-148,-147,-148,-147,-147,-147,-147,-146,-146,
-146,-146,-145,-145,-145,-144,-144,-144,-144,-143,-142,-142,-142,-142,-141,-140,-140,-140,-139,-139,
-138,-138,-138,-136,-137,-136,-135,-135,-134,-133,-133,-133,-132,-131,-131,-130,-129,-129,-129,-127,
-127,-127,-125,-125,-125,-123,-123,-123,-121,-121,-120,-120,-119,-118,-117,-116,-116,-115,-115,-113,
-113,-112,-112,-110,-110,-109,-108,-107,-107,-106,-105,-104,-103,-103,-102,-101,-100, -99, -99, -97,
 -97, -96, -95, -94, -94, -93, -91, -91, -90, -90, -88, -88, -86, -86, -85, -84, -84, -82, -82, -81,
 -80, -79, -78, -77, -77, -75, -75, -74, -73, -73, -71, -71, -69, -69, -68, -67, -67, -65, -65, -64,
 -63, -62, -62, -60, -60, -59, -58, -57, -57, -55, -55, -54, -53, -53, -52, -50, -50, -50, -48, -48,
 -46, -46, -46, -44, -44, -43, -43, -41, -41, -40, -40, -39, -38, -37, -36, -36, -35, -35, -33, -34,
 -32, -32, -30, -31, -29, -29, -29, -27, -27, -27, -25, -26, -24, -24, -23, -23, -22, -21, -21, -20,
 -20, -19, -18, -18, -18, -16, -17, -15, -15, -15, -14, -14, -13, -12, -12, -12, -11, -10, -11,  -9,
  -9,  -9,  -8,  -8,  -7,  -7,  -6,  -6,  -6,  -5,  -5,  -4,  -4,  -4,  -3,  -2,  -3,  -2,  -1,  -2,
  -1,   0,   0,   0,   0,   1,   1,   2,   1,   2,   3,   3,   2,   4,   3,   4,   4,   4,   5,   5,
   5,   5,   6,   5,   7,   6,   6,   7,   7,   7,   7,   8,   7,   8,   8,   8,   9,   8,   9,   9,
   9,   9,   9,   9,  10,  10,  10,   9,  10,  11,  10,  10,  11,  10,  11,  11,  10,  11,  11,  11,
  11,  12,  11,  11,  12,  11,  11,  12,  11,  12,  12,  11,  12,  12,  11,  12,  12,  12,  11,  12,
  12,  12,  12,  11,  12,  12,  12,  11,  12,  12,  12,  11,  12,  12,  11,  12,  11,  12,  11,  12,
  11,  11,  12,  11,  11,  11,  11,  11,  11,  11,  11,  11,  10,  11,  11,  10,  11,  10,  11,  10,
  10,  10,  10,  10,  10,  10,  10,   9,  10,   9,  10,   9,   9,  10,   9,   9,   9,   8,   9,   9,
   9,   8,   8,   9,   8,   8,   8,   8,   8,   8,   8,   7,   8,   7,   7,   8,   7,   7,   7,   7,
   7,   6,   7,   7,   6,   6,   7,   6,   6,   6,   6,   6,   5,   6,   6,   5,   6,   5,   5,   5,
   5,   5,   5,   5,   5,   5,   4,   5,   4,   4,   5,   4,   4,   4,   4,   4,   4,   3,   4,   3,
   4,   3,   4,   3,   3,   3,   4,   3,   3,   2,   3,   3,   3,   2,   3,   2,   3,   2,   3,   2,
   2,   2,   2,   2,   2,   2,   2,   2,   2,   1,   2,   2,   1,   2,   1,   2,   1,   1,   2,   0,
   1,   1,   2,   1,   1,   1,   1,   1,   0,   1,   1,   1,   1,   0,   1,   1,   0,   1,   0,   1,
   1,   0,   0,   1,   0,   1,   0,   0,   1,   0,   0,   0,   0,   1,   0,   0,   0,   0,   0,   0,
   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  -1,   0,   0,   0,   0,
   0,  -1,   0,   0,   0,   0,  -1,   0,   0,   0,  -1,   0,   0,   0,  -1,   0,   0,  -1,   0,   0,
   0,  -1,   0,   0,  -1,   0,   0,   0,  -1,   0,   0,  -1,   0,   0,   0,  -1,   0,   0,  -1,   0,
   0,   0,  -1,   0,   0,   0,  -1,   0,   0,   0,  -1,   0,   0,   0,  -1,   0,   0,   0,  -1,   0,
   0,   0,   0,  -1,   0,   0,   0,  -4,};
/* *INDENT-ON* */


static inline int32_t
FilterUp (int16_t Imp[], int16_t ImpD[], uint16_t Nwing, char Interp,
	int16_t * Xp, int16_t Ph, int16_t Inc)
{
	int16_t *Hp, *Hdp = NULL, *End;
	int16_t a = 0;
	int32_t v, t;

	v = 0;
	Hp = &Imp[Ph >> Na];
	End = &Imp[Nwing];
	if (Interp)
	{
		Hdp = &ImpD[Ph >> Na];
		a = Ph & Amask;
	}
	if (Inc == 1)										   /* If doing right wing...              */
	{													   /* ...drop extra coeff, so when Ph is  */
		End--;											   /*    0.5, we don't do too many mult's */
		if (Ph == 0)									   /* If the phase is zero...           */
		{												   /* ...then we've already skipped the */
			Hp += Npc;									   /*    first sample, so we must also  */
			Hdp += Npc;									   /*    skip ahead in Imp[] and ImpD[] */
		}
	}
	if (Interp)
		while (Hp < End)
		{
			t = *Hp;									   /* Get filter coeff */
			t += (((int32_t) * Hdp) * a) >> Na;			   /* t is now interp'd filter coeff */
			Hdp += Npc;									   /* Filter coeff differences step */
			t *= *Xp;									   /* Mult coeff by input sample */
			if (t & (1 << (Nhxn - 1)))					   /* Round, if needed */
				t += (1 << (Nhxn - 1));
			t >>= Nhxn;									   /* Leave some guard bits, but come back some */
			v += t;										   /* The filter output */
			Hp += Npc;									   /* Filter coeff step */
			Xp += Inc;									   /* Input signal step. NO CHECK ON BOUNDS */
		}
	else
		while (Hp < End)
		{
			t = *Hp;									   /* Get filter coeff */
			t *= *Xp;									   /* Mult coeff by input sample */
			if (t & (1 << (Nhxn - 1)))					   /* Round, if needed */
				t += (1 << (Nhxn - 1));
			t >>= Nhxn;									   /* Leave some guard bits, but come back some */
			v += t;										   /* The filter output */
			Hp += Npc;									   /* Filter coeff step */
			Xp += Inc;									   /* Input signal step. NO CHECK ON BOUNDS */
		}
	return (v);
}


static inline int32_t
FilterUD (int16_t Imp[], int16_t ImpD[], uint16_t Nwing, char Interp,
	int16_t * Xp, int16_t Ph, int16_t Inc, uint16_t dhb)
{
	int16_t a;
	int16_t *Hp, *Hdp, *End;
	int32_t v, t;
	uint32_t Ho;

	v = 0;
	Ho = (Ph * (uint32_t) dhb) >> Np;
	End = &Imp[Nwing];
	if (Inc == 1)										   /* If doing right wing...              */
	{													   /* ...drop extra coeff, so when Ph is  */
		End--;											   /*    0.5, we don't do too many mult's */
		if (Ph == 0)									   /* If the phase is zero...           */
			Ho += dhb;									   /* ...then we've already skipped the */
	}													   /*    first sample, so we must also  */
	/*    skip ahead in Imp[] and ImpD[] */
	if (Interp)
		while ((Hp = &Imp[Ho >> Na]) < End)
		{
			t = *Hp;									   /* Get IR sample */
			Hdp = &ImpD[Ho >> Na];						   /* get interp (lower Na) bits from diff table */
			a = Ho & Amask;								   /* a is logically between 0 and 1 */
			t += (((int32_t) * Hdp) * a) >> Na;			   /* t is now interp'd filter coeff */
			t *= *Xp;									   /* Mult coeff by input sample */
			if (t & 1 << (Nhxn - 1))					   /* Round, if needed */
				t += 1 << (Nhxn - 1);
			t >>= Nhxn;									   /* Leave some guard bits, but come back some */
			v += t;										   /* The filter output */
			Ho += dhb;									   /* IR step */
			Xp += Inc;									   /* Input signal step. NO CHECK ON BOUNDS */
		}
	else
		while ((Hp = &Imp[Ho >> Na]) < End)
		{
			t = *Hp;									   /* Get IR sample */
			t *= *Xp;									   /* Mult coeff by input sample */
			if (t & 1 << (Nhxn - 1))					   /* Round, if needed */
				t += 1 << (Nhxn - 1);
			t >>= Nhxn;									   /* Leave some guard bits, but come back some */
			v += t;										   /* The filter output */
			Ho += dhb;									   /* IR step */
			Xp += Inc;									   /* Input signal step. NO CHECK ON BOUNDS */
		}
	return (v);
}


static inline int16_t
WordToHword (int32_t v, int scl)
{
	int16_t out;
	int32_t llsb = (1 << (scl - 1));

	v += llsb;											   /* round */
	v >>= scl;
	if (v > SHRT_MAX)
		v = SHRT_MAX;
	else if (v < SHRT_MIN)
		v = SHRT_MIN;
	out = (int16_t) v;
	return out;
}


static int
SrcUp (struct rate_private_data *data, int voice, int16_t X[], char *dst_ptr,
	uint64_t * Time, uint32_t Nx, uint32_t Ny, uint16_t Nwing, uint16_t LpScl,
	int16_t Imp[], int16_t ImpD[], char Interp)
{
	int16_t *Xp;
	int32_t v;
	int32_t ocount;

	uint32_t dtb;						/* Fixed-point version of Dt */
	uint64_t endTime;					/* When Time reaches EndTime, return to user */

#ifdef USE_FPU
	double  factor;						/* factor = outSampleRate/inSampleRate */

	factor = 1.0 * BITS / data->pitch;
	double  dt;							/* Step through input signal */

	dt = 1.0 / factor;									   /* Output sampling period */
	dtb = dt * (1 << Np) + 0.5;							   /* Fixed-point representation */
#else
	dtb = (data->pitch * (1 << Np) + BITS / 2) / BITS;
#endif

	ocount = 0;
	endTime = *Time + (1 << Np) * (uint64_t) Nx;
	while (*Time < endTime)
	{
		Xp = &X[*Time >> Np];							   /* Ptr to current input sample */
		/* Perform left-wing inner product */
		v = FilterUp (Imp, ImpD, Nwing, Interp, Xp, (int16_t) (*Time & Pmask), -1);
		/* Perform right-wing inner product */
		v += FilterUp (Imp, ImpD, Nwing, Interp, Xp + 1,
			/* previous (triggers warning): (int16_t)((-*Time)&Pmask),1); */
			(int16_t) ((((*Time) ^ Pmask) + 1) & Pmask), 1);
		v >>= Nhg;										   /* Make guard bits */
		v *= LpScl;										   /* Normalize for unity filter gain */
		data->put (dst_ptr + ocount++ * data->src_voices * data->sample_size +
			voice * data->sample_size, WordToHword (v, NLpScl));
		if (ocount >= Ny)
			break;
		*Time += dtb;									   /* Move to next sample by time increment */
	}
	return (ocount);									   /* Return the number of output samples */
}


static int
SrcUD (struct rate_private_data *data, int voice,
	int16_t X[], char *dst_ptr, uint64_t * Time, uint32_t Nx, uint32_t Ny,
	uint16_t Nwing, uint16_t LpScl, int16_t Imp[], int16_t ImpD[], char Interp)
{
	int16_t *Xp;
	int32_t v;
	int32_t ocount;

	uint64_t endTime;					/* When Time reaches EndTime, return to user */
	uint32_t dhb, dtb;					/* Fixed-point versions of Dh,Dt */

#ifdef USE_FPU
	double  factor;						/* factor = outSampleRate/inSampleRate */
	double  dh;							/* Step through filter impulse response */
	double  dt;							/* Step through input signal */

	factor = 1.0 * BITS / data->pitch;
	dt = 1.0 / factor;									   /* Output sampling period */
	dtb = dt * (1 << Np) + 0.5;							   /* Fixed-point representation */
	dh = min (Npc, factor * Npc);						   /* Filter sampling period */
	dhb = dh * (1 << Na) + 0.5;							   /* Fixed-point representation */
#else
	dtb = (data->pitch * (1 << Np) + BITS / 2) / BITS;
	dhb = min (Npc * (1 << Na),
		(((uint64_t) 1 << (2 * SHIFT)) / data->pitch * Npc * (1 << Na) + BITS / 2) / BITS);
#endif


	ocount = 0;
	endTime = *Time + (1 << Np) * (uint64_t) Nx;
	while (*Time < endTime)
	{
		Xp = &X[*Time >> Np];							   /* Ptr to current input sample */
		v = FilterUD (Imp, ImpD, Nwing, Interp, Xp, (int16_t) (*Time & Pmask), -1, dhb);	/* Perform left-wing inner product */
		v += FilterUD (Imp, ImpD, Nwing, Interp, Xp + 1,
			/* previous (triggers warning): (int16_t)((-*Time)&Pmask), */
			(int16_t) ((((*Time) ^ Pmask) + 1) & Pmask), 1, dhb);	/* Perform right-wing inner product */
		v >>= Nhg;										   /* Make guard bits */
		v *= LpScl;										   /* Normalize for unity filter gain */
		data->put (dst_ptr + ocount++ * data->src_voices * data->sample_size +
			voice * data->sample_size, WordToHword (v, NLpScl));
		if (ocount >= Ny)
			break;
		*Time += dtb;									   /* Move to next sample by time increment */
	}
	return (ocount);									   /* Return the number of output samples */
}

/************************************************************************/
/* END BSD License                                                      */
/************************************************************************/


static void
resample_band_limited (struct rate_private_data *data, int voices,
	char *src_ptr, int src_size, char *dst_ptr, int dst_size)
{
	uint16_t Nmult;						/* Filter length for up-conversions */
	int16_t *Imp = 0;					/* Filter coefficients */
	int16_t *ImpD = 0;					/* ImpD[n] = Imp[n+1]-Imp[n] */
	uint16_t LpScl;						/* Unity-gain scale factor */
	uint16_t Nwing;						/* Filter table size */

	uint64_t Time;						/* Current time/pos in input sample */

	uint32_t Xoff;
	uint32_t Nout, Nx;
	int     voice;
	int     count;
	char   *src;
	int     src_inc;

#ifdef USE_FPU
	double  factor;						/* factor = outSampleRate/inSampleRate */
#endif

	Nmult = data->Nmult;
	Imp = data->Imp;
	ImpD = data->ImpD;
	LpScl = data->LpScl;
	Nwing = data->Nwing;

#ifdef USE_FPU
	factor = 1.0 * BITS / data->pitch;
	Xoff = ((Nmult + 1) / 2.0) * max (1.0, 1.0 / factor) + 10;
#else
	Xoff = (Nmult + 1) * max ((1 << SHIFT), data->pitch) / BITS / 2 + 10;
#endif

#ifdef USE_FPU
	if (factor < 1)
		LpScl = LpScl * factor + 0.5;
#else
	if (data->pitch >= (1 << SHIFT))
		LpScl = ((uint64_t) LpScl * BITS + data->pitch / 2) / data->pitch;
#endif

	for (voice = 0; voice < voices; ++voice)
	{
		src = src_ptr + voice * data->sample_size;
		src_inc = voices * data->sample_size;
		for (count = 0; count < src_size; count++)
		{
			data->X[voice][2 * Xoff + count] = data->take (src);
			src += src_inc;
		}

		Nx = src_size;
		Time = 0;
		Time = (Xoff << Np);							   /* Current-time pointer for converter */

#ifdef USE_FPU
		if (factor >= 1)
#else
		if (data->pitch < (1 << SHIFT))
#endif
		{												   /* SrcUp() is faster if we can use it */
			Nout = SrcUp (data, voice, data->X[voice], dst_ptr, &Time, Nx, dst_size,
				Nwing, LpScl, Imp, ImpD, USE_INTERP);
		}
		else
		{
			Nout = SrcUD (data, voice, data->X[voice], dst_ptr, &Time, Nx, dst_size,
				Nwing, LpScl, Imp, ImpD, USE_INTERP);
		}

		for (count = 0; count < 2 * Xoff; count++)
		{
			data->X[voice][count] = data->X[voice][count + Nx];
		}
	}
}


static void
resample_linear (struct rate_private_data *data, int voices,
	char *src_ptr, int src_size, char *dst_ptr, int dst_size)
{
	int64_t pos;
	int     voice;
	int     count;
	signed int a;
	char   *src;
	signed short S1, S2;
	signed int val;
	char   *dst;

	for (voice = 0; voice < voices; ++voice)
	{
		pos = 0;
		for (count = 0; count < dst_size; count++)
		{
			a = (pos) >> SHIFT;
			if ((a - 1) >= 0)
			{
				src = src_ptr + (a - 1) * voices * data->sample_size + voice * data->sample_size;
				S1 = data->take (src);
			}
			else
				S1 = data->S[voice];
			src = src_ptr + (a + 0) * voices * data->sample_size + voice * data->sample_size;
			S2 = data->take (src);

			val = S1 + ((S2 - S1) * (pos & MASK) / BITS);

			if (val < SHRT_MIN)
				val = SHRT_MIN;
			else if (val > SHRT_MAX)
				val = SHRT_MAX;

			dst = dst_ptr + count * voices * data->sample_size + voice * data->sample_size;
			data->put (dst, val);

			pos += data->pitch;
		}
		src = src_ptr + (src_size - 1) * voices * data->sample_size + voice * data->sample_size;
		data->S[voice] = data->take (src);
	}
}


static  ssize_t
rate_src_size (snd_pcm_plugin_t * plugin, size_t size)
{
	struct rate_private_data *data;
	int64_t dst_fs, src_fs;

	if (plugin == NULL || size <= 0)
		return -EINVAL;
	data = (struct rate_private_data *) snd_pcm_plugin_extra_data (plugin);

	if (data->src_mode == SND_SRC_MODE_NORMAL)
	{
		if (data->old_dst_size == size)
			return (data->old_src_size);

		dst_fs = size;
		src_fs = ((((int64_t) size * data->pitch) + (BITS / 2)) >> SHIFT);
		src_fs = src_fs / (data->src_voices * data->sample_size) *
			(data->src_voices * data->sample_size);
		data->pitch = ((int64_t) src_fs << SHIFT) / dst_fs;

		data->old_dst_size = dst_fs;
		data->old_src_size = src_fs;
	}
	else /* variable src frag size */
	{
		dst_fs = size;
		if (data->newfragsize == 0)
		{
			src_fs = ((((int64_t) size * data->pitch) + (BITS / 2)) >> SHIFT);
			src_fs = src_fs / (data->src_voices * data->sample_size) *
				(data->src_voices * data->sample_size);
		}
		else
			src_fs = data->newfragsize;
		data->pitch = ((int64_t) src_fs << SHIFT) / dst_fs;

		data->old_dst_size = dst_fs;
		data->old_src_size = src_fs;
	}
	return (src_fs);
}

static  ssize_t
rate_dst_size (snd_pcm_plugin_t * plugin, size_t size)
{
	struct rate_private_data *data;
	int64_t dst_fs, src_fs;

	if (plugin == NULL || size <= 0)
		return -EINVAL;
	data = (struct rate_private_data *) snd_pcm_plugin_extra_data (plugin);

	if (data->old_src_size == size)
		return (data->old_dst_size);

	/* adjust the pitch for integer fragsizes at src & dst */
	src_fs = size;
	if (data->src_mode == SND_SRC_MODE_NORMAL)
	{

		dst_fs = ((((int64_t) src_fs << SHIFT) + (data->pitch / 2)) / data->pitch);
		dst_fs = dst_fs / (data->dst_voices * data->sample_size)
			* (data->dst_voices * data->sample_size);
	}
	else
		dst_fs = data->dstfragsize;
	data->pitch = ((int64_t) src_fs << SHIFT) / dst_fs;

	data->old_dst_size = dst_fs;
	data->old_src_size = src_fs;
	return (dst_fs);
}

static int
rate_init (snd_pcm_plugin_t * plugin)
{
	struct rate_private_data *data;
	int     i;
	uint16_t extra;
	
	data = (struct rate_private_data *) snd_pcm_plugin_extra_data (plugin);
	
	if ((data->src_mode == SND_SRC_MODE_ASYNC) && !asrc_conf_done) 
	{
		char *str;
		asrc_conf_done = 1;

		str = getenv("MAXPHASE");
		if (str) {
			asrc_maxphase = strtof(str, NULL);
		}

		str = getenv("MINPHASE");
		if (str) {
			asrc_minphase = strtof(str, NULL);
		}

		str = getenv("ASRCTC");
		if (str) {
			asrc_tc = strtof(str, NULL);
		}

		str = getenv("ASRCADJ");
		if (str) {
			asrc_adj = strtol(str, NULL, 10);
		}
		
		str = getenv("ASRCDAMP");
		if (str) {
			asrc_damp = strtol(str, NULL, 10);
		}
		
		str = getenv("ASRCDEBUG");
		if (str) {
			asrc_debug = strtol(str, NULL, 10);
		}
		
		if (asrc_debug)
		{
			slogf (_SLOGC_AUDIO, _SLOG_ERROR, "ASRC Using values: MAXPHASE: %f, MINPHASE: %f, ASRCTC: %f, ASRCADJ: %d, ASRCDAMP: %d, ASRCDEBUG: %d",
				asrc_maxphase, asrc_minphase, asrc_tc, asrc_adj, asrc_damp, asrc_debug);
		}
	}
			
	for (i = 0; i < data->src_voices; i++)
	{
		if (data->X[i])
			free (data->X[i]);
	}

	if (data->X[0] == NULL && data->method == resample_band_limited)
	{
		extra = ((data->Nmult + 1) * 8 / 2) + 10;
		for (i = 0; i < data->src_voices; i++)
		{
			if ((data->X[i] = malloc (data->old_src_size * sizeof (int16_t)
						+ extra * 2 + 10)) == NULL)
				return (-ENOMEM);
		}
	}
	return (0);
}

static void
rate_prepare (snd_pcm_plugin_t * plugin)
{
	struct rate_private_data *data;
	unsigned char silence;
	int     i;
	uint16_t extra;

	data = (struct rate_private_data *) snd_pcm_plugin_extra_data (plugin);
	silence = snd_pcm_plugin_silence (&data->format);
	if (data->method == resample_band_limited)
	{
		for (i = 0; i < data->src_voices; i++)
		{
			extra = ((data->Nmult + 1) * 8 / 2) + 10;
			if (data->X[i])
				memset (data->X[i], silence, data->old_src_size *
					sizeof (int16_t) + extra * 2 + 10);
		}
	}
	else
	{
		for (i = 0; i < data->src_voices; i++)
		{
			memset (&data->S[i], silence, sizeof (data->S[i]));
		}
	}
}

static  ssize_t
rate_transfer (snd_pcm_plugin_t * plugin,
	void *src_ptr, size_t src_size, void *dst_ptr, size_t dst_size)
{
	struct rate_private_data *data;
	int     rtn;

	if (plugin == NULL || src_ptr == NULL || src_size < 0 || dst_ptr == NULL || dst_size < 0)
		return -EINVAL;
	if (src_size == 0)
		return 0;
	data = (struct rate_private_data *) snd_pcm_plugin_extra_data (plugin);
	if (data == NULL)
		return -EINVAL;

	if (src_size != data->old_src_size || dst_size != data->old_dst_size)
	{
		data->pitch = ((int64_t) src_size << SHIFT) / dst_size;
		data->old_src_size = src_size;
		data->old_dst_size = dst_size;
		if ((rtn = rate_init (plugin)) != 0)
			return (rtn);
		rate_prepare (plugin);
	}

	data->method (data, data->src_voices,
		src_ptr, src_size / (data->src_voices * data->sample_size),
		dst_ptr, dst_size / (data->dst_voices * data->sample_size));

	data->icount += src_size;
	data->ocount += dst_size;

	return rate_dst_size (plugin, src_size);
}



static int
rate_action (snd_pcm_plugin_t * plugin, snd_pcm_plugin_action_t action)
{
	struct rate_private_data *data;
	int     rtn = 0;

	if (plugin == NULL)
		return -EINVAL;
	data = (struct rate_private_data *) snd_pcm_plugin_extra_data (plugin);
	switch (action)
	{
	case INIT:
		rtn = rate_init (plugin);
		break;
	case PREPARE:
		rate_prepare (plugin);
		break;
	case DRAIN:
		break;
	case FLUSH:
		break;
	}
	return rtn;
}

static void
rate_free (snd_pcm_plugin_t * plugin, void *private_data)
{
	struct rate_private_data *data;
	int     i;

	if (plugin == NULL)
		return;
	data = (struct rate_private_data *) snd_pcm_plugin_extra_data (plugin);

	for (i = 0; i < data->src_voices; i++)
	{
		if (data->X[i])
			free (data->X[i]);
	}
}

int
snd_pcm_plugin_build_rate (snd_pcm_format_t * src_format,
	snd_pcm_format_t * dst_format, unsigned int method,
	snd_pcm_plugin_t ** r_plugin, unsigned int src_mode,
	int target, int32_t out_frag_size)
{
	struct rate_private_data *data;
	snd_pcm_plugin_t *plugin;

	if (r_plugin == NULL)
		return -EINVAL;
	*r_plugin = NULL;

	if (src_format->interleave != dst_format->interleave && src_format->voices > 1)
		return -EINVAL;
	if (src_format->format != dst_format->format)
		return -EINVAL;
	if (!dst_format->interleave && src_format->voices > 1)
		return -EINVAL;
	if (src_format->voices != dst_format->voices)
		return -EINVAL;
	if (dst_format->voices < 1 || dst_format->voices > MAX_VOICES)
		return -EINVAL;

	if (src_format->format > SND_PCM_SFMT_S32_BE)
		return -EINVAL;
	if (src_format->rate == dst_format->rate && src_mode == SND_SRC_MODE_NORMAL)
		return -EINVAL;
	plugin = snd_pcm_plugin_build ("rate conversion", sizeof (struct rate_private_data));
	if (plugin == NULL)
		return -ENOMEM;
	data = (struct rate_private_data *) snd_pcm_plugin_extra_data (plugin);
	data->format = *src_format;
	data->src_voices = src_format->voices;
	data->dst_voices = dst_format->voices;
	data->src_rate = src_format->rate;
	data->dst_rate = dst_format->rate;
	data->sample_size = snd_pcm_format_width (src_format->format) / 8;
	if (method == 1)
	{
		data->method = resample_band_limited;
	}
	else
	{
		data->method = resample_linear;
	}
	data->take = rate_take_sample[src_format->format];
	data->put = rate_put_sample[dst_format->format];
	data->pitch = (((uint64_t) src_format->rate << SHIFT) +
		(dst_format->rate >> 1)) / dst_format->rate;
	data->old_src_size = data->old_dst_size = 0;
	data->src_mode = SND_SRC_MODE_NORMAL;
	if (src_mode > SND_SRC_MODE_NORMAL)
	{
		data->src_mode = src_mode;
		data->src_ratio = (double)data->src_rate / data->dst_rate;
		data->target = target;
		data->filttarget = data->target;
		data->src_offset = 0;
		data->icount = 0;
		data->ocount = 0;
		data->newfragsize = 0;
		data->fractfrag = 0;
		data->dstfragsize = out_frag_size;
		data->initialized = 0;
	}

	/* band limited specific init */
	data->Nmult = A_FILTER_NMULT;
	data->Imp = A_FILTER_IMP;							   /* Impulse response */
	data->ImpD = A_FILTER_IMPD;							   /* Impulse response deltas */
	data->LpScl = A_FILTER_SCALE;						   /* Unity-gain scale factor */
	data->Nwing = A_FILTER_NWING;						   /* Filter table length */

	plugin->transfer = rate_transfer;
	plugin->src_size = rate_src_size;
	plugin->dst_size = rate_dst_size;
	plugin->action = rate_action;
	plugin->private_free = rate_free;
	*r_plugin = plugin;
	return 0;
}


unsigned int
snd_pcm_plugin_set_src_method (snd_pcm_t * pcm, unsigned int method)
{
	if (method <= 1)
		pcm->plugin_src_method = method;
	return (pcm->plugin_src_method);
}

unsigned int
snd_pcm_plugin_set_src_mode (snd_pcm_t * pcm, unsigned int src_mode, int target)
{
	if (src_mode <= SND_SRC_MODE_ASYNC)
	{
		pcm->plugin_src_mode = src_mode;
		pcm->plugin_src_target = target;
	}
	return (pcm->plugin_src_mode);
}

int
snd_pcm_plugin_src_max_frag (snd_pcm_t * pcm, unsigned int fragsize)
{
	struct rate_private_data *data;
	snd_pcm_plugin_t *plugin;
	int extrafrag;

	plugin = snd_pcm_plugin_find (pcm, 0, "rate conversion");
	if (plugin == NULL)
		return -EINVAL;
	data = (struct rate_private_data *) snd_pcm_plugin_extra_data (plugin);

	if (!data->initialized)
	{
		extrafrag = data->src_voices * data->sample_size;
		if (data->src_mode == SND_SRC_MODE_ACTUAL)
		{
			data->maxfragsize = fragsize + extrafrag;
			data->minfragsize = fragsize;
		}
		else if (data->src_mode == SND_SRC_MODE_ASYNC)
		{
			data->maxfragsize = (fragsize + extrafrag) * MAXPHASE;
			data->minfragsize = fragsize * MINPHASE;
		}
		else
		{
			data->maxfragsize = fragsize;
			data->minfragsize = fragsize;
		}
		data->initialized = 1;
	}
	return (data->maxfragsize);
}

int
snd_pcm_plugin_update_src (snd_pcm_t * pcm, snd_pcm_channel_setup_t * setup, int currlevel)
{
	struct rate_private_data *data;
	snd_pcm_plugin_t *plugin;
	double truecount;
	int extrafrag;

	plugin = snd_pcm_plugin_find (pcm, 0, "rate conversion");
	if (plugin == NULL)
		return -EINVAL;
	data = (struct rate_private_data *) snd_pcm_plugin_extra_data (plugin);

	/* In NORMAL mode return without changing the src size */
	if (data->src_mode == SND_SRC_MODE_NORMAL)
		return data->old_src_size;

	/* ASRC compensation - adjusts src ratio */
	if (data->src_mode == SND_SRC_MODE_ASYNC)
	{
		data->filttarget = ASRCTC * data->filttarget + (1 - ASRCTC) * currlevel;
		if ((data->filttarget - data->target) > ASRCADJ) {
			if (currlevel > data->target && data->src_offset < 0) {
				data->src_offset += ASRCDAMP;
			} else {
				data->src_offset += 1;
			}
		} else if ((data->target - data->filttarget) > ASRCADJ) {
			if (currlevel < data->target && data->src_offset > 0) {
				data->src_offset -= ASRCDAMP;
			} else {
				data->src_offset -= 1;
			}
		}

		data->src_ratio = (double)(data->src_rate + data->src_offset) / data->dst_rate;

		if (asrc_debug) {
			slogf (_SLOGC_AUDIO, _SLOG_ERROR, "ASRC: target=%i, filttarget=%f, src_ratio=%f, src_offset=%i",
				data->target, data->filttarget, data->src_ratio, data->src_offset);
		}
	}

	/* True rate - adjusts frag source size */
	truecount = data->ocount * data->src_ratio + data->fractfrag;
	extrafrag = data->src_voices * data->sample_size;
	data->newfragsize = data->old_src_size;
	if ((truecount - extrafrag) > data->icount && (data->old_src_size + extrafrag) <= data->maxfragsize)
	{
		data->newfragsize = data->old_src_size + extrafrag;
		data->fractfrag = truecount - data->icount - extrafrag;
		data->icount = 0;
		data->ocount = 0;
		pdprintf ("adding a sample\n");
	}
	else if (data->icount > truecount && truecount != 0 && (data->old_src_size - extrafrag) >= data->minfragsize)
	{
		data->newfragsize = data->old_src_size - extrafrag;
		data->fractfrag = truecount - data->icount + extrafrag;
		data->icount = 0;
		data->ocount = 0;
		pdprintf("dropping a sample\n");
	}

	/* Prevent the counters from wrapping */
	if (data->src_mode == SND_SRC_MODE_ASYNC)
	{
		if ((data->old_src_size + extrafrag >= data->maxfragsize || data->old_src_size - extrafrag <= data->minfragsize) &&
				truecount > CNTRESET )
		{
			data->icount = 0;
			data->ocount = 0;
			data->fractfrag = 0;
		}
	}
	else if (data->icount == data->ocount)
	{
		data->icount = 0;
		data->ocount = 0;
	}

	if (asrc_debug) {
		slogf (_SLOGC_AUDIO, _SLOG_ERROR, "ASRC: truecount=%f, icount=%i, ocount=%i, ratio=%f, newfragsize=%i, lvl=%d",
			truecount, data->icount, data->ocount, data->src_ratio, data->newfragsize, currlevel);
	}

	if (data->newfragsize != data->old_src_size)
		snd_pcm_plugin_setup (pcm, setup);

	return data->newfragsize;
}
