/*
 *  Voices conversion Plug-In
 *  Copyright (c) 1999 by Jaroslav Kysela <perex@suse.cz>
 *
 *
 *   This library is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Library General Public License as
 *   published by the Free Software Foundation; either version 2 of
 *   the License, or (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU Library General Public License for more details.
 *
 *   You should have received a copy of the GNU Library General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <endian.h>
#include <byteswap.h>
#include "../pcm_local.h"

__SRCVERSION( "$URL: http://svn.ott.qnx.com/product/branches/6.5.0/SP1/lib/asound/pcm/plugin/voices.c $ $Rev: 648519 $" )

/*
 *  Basic voices conversion plugin
 */

struct voices_private_data
{
	int     src_voices;
	int     dst_voices;
	int     width;						/* in bites */
	int     flg_signed:1;
	int     flg_endian:1;				/* cpu different endian */
	int     channel;
	uint32_t voice_matrix[32];
};

#define CONVERT_VOICES_FUNC(name, type) \
inline void convert_voices_##name(int channel, type *src_ptr, type *dst_ptr, size_t src_buf_size, int src_voices,	\
									 int dst_voices, uint32_t voice_matrix[])		\
{																					\
	int dst_bidx, src_bidx, sample_cnt, src_vidx, dst_vidx;							\
	int *app_vidx = (channel == SND_PCM_CHANNEL_PLAYBACK ? &src_vidx : &dst_vidx);	\
	int *hw_vidx = (channel == SND_PCM_CHANNEL_PLAYBACK ? &dst_vidx : &src_vidx);	\
																					\
	for (dst_bidx=src_bidx=sample_cnt=0; sample_cnt < src_buf_size / src_voices;	\
			dst_bidx+=dst_voices, src_bidx+=src_voices, sample_cnt++)				\
	{																				\
		for (src_vidx = 0; src_vidx < src_voices; src_vidx++)						\
		{																			\
			for (dst_vidx = 0; dst_vidx < dst_voices; dst_vidx++)					\
			{																		\
				if ( (voice_matrix[*app_vidx] >> *hw_vidx) & 0x1)					\
					dst_ptr[dst_bidx + dst_vidx] = src_ptr[src_bidx + src_vidx];	\
				else if (src_vidx == 0)												\
					dst_ptr[dst_bidx + dst_vidx] = 0x0;								\
			}																		\
		}																			\
	}																				\
}

inline void convert_voices_24(int channel, uint8_t *src_ptr, uint8_t *dst_ptr, size_t src_buf_size, int src_voices,
									 int dst_voices, uint32_t voice_matrix[])
{
	int dst_bidx, src_bidx, sample_cnt, src_vidx, dst_vidx;	
	int *app_vidx = (channel == SND_PCM_CHANNEL_PLAYBACK ? &src_vidx : &dst_vidx);
	int *hw_vidx = (channel == SND_PCM_CHANNEL_PLAYBACK ? &dst_vidx : &src_vidx);

	for (dst_bidx=src_bidx=sample_cnt=0; sample_cnt < src_buf_size / src_voices;
			dst_bidx+=(3 * dst_voices), src_bidx+=(3 * src_voices), sample_cnt++)
	{
		for (src_vidx = 0; src_vidx < src_voices; src_vidx++)
		{
			for (dst_vidx = 0; dst_vidx < dst_voices; dst_vidx++)
			{
				if ( (voice_matrix[*app_vidx] >> *hw_vidx) & 0x1) {
					dst_ptr[dst_bidx + dst_vidx * 3] = src_ptr[src_bidx + src_vidx * 3];
					dst_ptr[dst_bidx + dst_vidx * 3 + 1] = src_ptr[src_bidx + src_vidx * 3 + 1];
					dst_ptr[dst_bidx + dst_vidx * 3 + 2] = src_ptr[src_bidx + src_vidx * 3 + 2];
				} else if (src_vidx == 0) {
					dst_ptr[dst_bidx + dst_vidx * 3] = 0x0;
					dst_ptr[dst_bidx + dst_vidx * 3 + 1] = 0x0;
					dst_ptr[dst_bidx + dst_vidx * 3 + 2] = 0x0;
			    }
			}
		}
	}
}


/* *INDENT-OFF* */
CONVERT_VOICES_FUNC (8, uint8_t)
CONVERT_VOICES_FUNC (16, uint16_t)
CONVERT_VOICES_FUNC (32, uint32_t)
/* *INDENT-ON* */

static  ssize_t
voices_transfer (snd_pcm_plugin_t * plugin,
	void *src_ptr, size_t src_size, void *dst_ptr, size_t dst_size)
{
	struct voices_private_data *data;

	if (plugin == NULL || src_ptr == NULL || src_size < 0 || dst_ptr == NULL || dst_size < 0)
		return -EINVAL;
	if (src_size == 0)
		return 0;
	data = (struct voices_private_data *) snd_pcm_plugin_extra_data (plugin);
	if (data == NULL)
		return -EINVAL;

	switch (data->width)
	{
	case 8:
		convert_voices_8 (data->channel, src_ptr, dst_ptr, src_size, data->src_voices,
			data->dst_voices, data->voice_matrix);
		break;

	case 16:
		convert_voices_16 (data->channel, src_ptr, dst_ptr, src_size / 2,
			data->src_voices, data->dst_voices, data->voice_matrix);
		break;

	case 24:
		convert_voices_24 (data->channel, (uint8_t*) src_ptr, (uint8_t*) dst_ptr, src_size / 3,
			data->src_voices, data->dst_voices, data->voice_matrix);
		break;		

	case 32:
		convert_voices_32 (data->channel, (uint32_t *) src_ptr, (uint32_t *) dst_ptr, src_size / 4,
			data->src_voices, data->dst_voices, data->voice_matrix);
		break;

	default:
		return -EINVAL;
	}
	return (src_size * data->dst_voices) / data->src_voices;
}

static  ssize_t
voices_src_size (snd_pcm_plugin_t * plugin, size_t size)
{
	struct voices_private_data *data;

	if (plugin == NULL || size <= 0)
		return -EINVAL;
	data = (struct voices_private_data *) snd_pcm_plugin_extra_data (plugin);
	return (size * data->src_voices) / data->dst_voices;
}

static  ssize_t
voices_dst_size (snd_pcm_plugin_t * plugin, size_t size)
{
	struct voices_private_data *data;

	if (plugin == NULL || size <= 0)
		return -EINVAL;
	data = (struct voices_private_data *) snd_pcm_plugin_extra_data (plugin);
	return (size * data->dst_voices) / data->src_voices;
}

int
snd_pcm_plugin_get_voice_conversion (snd_pcm_t * pcm, int channel,
	snd_pcm_voice_conversion_t * voice_conversion)
{
	snd_pcm_plugin_t *plugin;

	if (pcm == NULL || voice_conversion == NULL || channel < 0 || channel > 1)
		return -EINVAL;

	for (plugin = pcm->plugin_first[channel]; plugin != NULL; plugin = plugin->next)
	{
		if (strcmp (plugin->name, "voices conversion") == 0)
		{
			struct voices_private_data *data =
				(struct voices_private_data *) snd_pcm_plugin_extra_data (plugin);
			if (channel == SND_PCM_CHANNEL_PLAYBACK)
			{
				voice_conversion->app_voices = data->src_voices;
				voice_conversion->hw_voices = data->dst_voices;
			}
			else
			{
				voice_conversion->app_voices = data->dst_voices;
				voice_conversion->hw_voices = data->src_voices;
			}
			memcpy (voice_conversion->matrix, data->voice_matrix,
				sizeof (voice_conversion->matrix));
			return EOK;
		}
	}
	return -ENOENT;
}

int
snd_pcm_plugin_set_voice_conversion (snd_pcm_t * pcm, int channel,
	snd_pcm_voice_conversion_t * voice_conversion)
{
	snd_pcm_plugin_t *plugin;

	if (pcm == NULL || voice_conversion == NULL || channel < 0 || channel > 1)
		return -EINVAL;

	for (plugin = pcm->plugin_first[channel]; plugin != NULL; plugin = plugin->next)
	{
		if (strcmp (plugin->name, "voices conversion") == 0)
		{
			struct voices_private_data *data =
				(struct voices_private_data *) snd_pcm_plugin_extra_data (plugin);
			memcpy (data->voice_matrix, voice_conversion->matrix, sizeof (data->voice_matrix));
			return EOK;
		}
	}
	return -ENOENT;
}

int
snd_pcm_plugin_build_voices (snd_pcm_format_t * src_format,
	snd_pcm_format_t * dst_format, int channel, snd_pcm_plugin_t ** r_plugin)
{
	snd_pcm_plugin_t *plugin;
	struct voices_private_data *data;
	int     i;
	int     j;

	if (r_plugin == NULL)
		return -EINVAL;
	*r_plugin = NULL;

	if (src_format->interleave != dst_format->interleave && src_format->voices > 1)
		return -EINVAL;
	if (!dst_format->interleave)
		return -EINVAL;
	if (src_format->format != dst_format->format)
		return -EINVAL;
	if (src_format->rate != dst_format->rate)
		return -EINVAL;
	if (src_format->voices == dst_format->voices)
		return -EINVAL;
	if (src_format->voices < 1 || src_format->voices > 32 ||
		dst_format->voices < 1 || dst_format->voices > 32)
		return -EINVAL;

	if (src_format->format < SND_PCM_SFMT_U8 || src_format->format > SND_PCM_SFMT_S32_BE)
	{
		if (src_format->format != SND_PCM_SFMT_MU_LAW && src_format->format != SND_PCM_SFMT_A_LAW)
			return -EINVAL;
	}
	plugin = snd_pcm_plugin_build ("voices conversion", sizeof (struct voices_private_data));
	if (plugin == NULL)
		return -ENOMEM;
	data = (struct voices_private_data *) snd_pcm_plugin_extra_data (plugin);

	data->channel = channel;
	if (src_format->voices <= dst_format->voices)
	{

		memset (data->voice_matrix, 0, sizeof (data->voice_matrix));
		/* For stereo and quad destination replicate all */
		if (dst_format->voices == 2 || dst_format->voices == 4)
		{ 
			int    *app_vidx = (channel == SND_PCM_CHANNEL_PLAYBACK ? &i : &j);
			int    *hw_vidx = (channel == SND_PCM_CHANNEL_PLAYBACK ? &j : &i);

			for (i = 0; i < src_format->voices; i++)
			{
				for (j = i; j < dst_format->voices; j += src_format->voices)
					data->voice_matrix[*app_vidx] |= 1 << *hw_vidx;
			}
		}
		else	/* Silence fill missing channels for all other destination */
		{
			for (i = 0; i < src_format->voices; i++)
				data->voice_matrix[i] |= 1 << i;
		}
	}
	else
	{
		memset (data->voice_matrix, 0, sizeof (data->voice_matrix));
		for (i = 0; i < dst_format->voices; i++)
			data->voice_matrix[i] |= 1 << i;
	}

	data->src_voices = src_format->voices;
	data->dst_voices = dst_format->voices;
	data->width = snd_pcm_format_width (src_format->format);
	data->flg_signed = snd_pcm_format_signed (src_format->format);
#if __BYTE_ORDER == __LITTLE_ENDIAN
	data->flg_endian = snd_pcm_format_big_endian (src_format->format);
#elif __BYTE_ORDER == __BIG_ENDIAN
	data->flg_endian = snd_pcm_format_little_endian (src_format->format);
#endif
	plugin->transfer = voices_transfer;
	plugin->src_size = voices_src_size;
	plugin->dst_size = voices_dst_size;
	*r_plugin = plugin;
	return 0;
}

#ifdef __KERNEL__
EXPORT_SYMBOL (snd_pcm_plugin_build_voices);
#endif
